"""Tests for classifier compatibility, especially those without predict_proba."""

import numpy as np
import pytest

from lmprobe import LinearProbe
from lmprobe.classifiers import (
    BUILTIN_CLASSIFIERS,
    CLASSIFICATION_CLASSIFIERS,
    build_classifier,
    validate_classifier,
)


class TestBuiltinClassifiers:
    """Tests for all built-in classifiers."""

    def test_all_builtin_classifiers_can_be_built(self):
        """Every builtin classifier can be instantiated."""
        for name in BUILTIN_CLASSIFIERS:
            clf = build_classifier(name)
            assert clf is not None
            assert hasattr(clf, "fit")
            assert hasattr(clf, "predict")

    def test_unknown_classifier_raises(self):
        """Unknown classifier name raises ValueError."""
        with pytest.raises(ValueError, match="Unknown classifier"):
            build_classifier("nonexistent_classifier")

    @pytest.mark.parametrize("name", ["logistic_regression", "svm", "sgd", "mass_mean", "lda"])
    def test_classifiers_with_predict_proba(self, name):
        """These classifiers support predict_proba."""
        clf = build_classifier(name)
        assert hasattr(clf, "predict_proba")

    def test_ridge_lacks_predict_proba(self):
        """RidgeClassifier does not have predict_proba."""
        clf = build_classifier("ridge")
        assert not hasattr(clf, "predict_proba")


class TestValidateClassifier:
    """Tests for classifier validation."""

    def test_warns_on_missing_predict_proba(self):
        """Warns when classifier lacks predict_proba."""
        clf = build_classifier("ridge")
        with pytest.warns(UserWarning, match="does not support predict_proba"):
            validate_classifier(clf)

    def test_no_warning_with_predict_proba(self):
        """No warning for classifiers with predict_proba."""
        import warnings

        clf = build_classifier("logistic_regression")
        # Should not warn about predict_proba
        with warnings.catch_warnings(record=True) as record:
            warnings.simplefilter("always")
            validate_classifier(clf)
        # Filter to only UserWarnings about predict_proba
        proba_warnings = [w for w in record if "predict_proba" in str(w.message)]
        assert len(proba_warnings) == 0

    def test_raises_on_missing_fit(self):
        """Raises TypeError if classifier lacks fit()."""
        class NoFit:
            def predict(self, X):
                pass

        with pytest.raises(TypeError, match="must have a fit"):
            validate_classifier(NoFit())

    def test_raises_on_missing_predict(self):
        """Raises TypeError if classifier lacks predict()."""
        class NoPredict:
            def fit(self, X, y):
                pass

        with pytest.raises(TypeError, match="must have a predict"):
            validate_classifier(NoPredict())


class TestLinearProbeWithRidge:
    """Tests for LinearProbe with RidgeClassifier (no predict_proba)."""

    def test_ridge_fit_works(self, tiny_model):
        """RidgeClassifier can be used for training."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            classifier="ridge",
            device="cpu",
            remote=False,
            random_state=42,
        )

        # Should succeed without error
        probe.fit(["positive one", "positive two"], ["negative one", "negative two"])
        assert probe.classifier_ is not None

    def test_ridge_predict_works(self, tiny_model):
        """predict() works with RidgeClassifier."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            classifier="ridge",
            device="cpu",
            remote=False,
            random_state=42,
        )

        probe.fit(["positive"], ["negative"])
        predictions = probe.predict(["test input"])

        assert predictions.shape == (1,)
        assert predictions[0] in [0, 1]

    def test_ridge_score_works(self, tiny_model):
        """score() works with RidgeClassifier."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            classifier="ridge",
            device="cpu",
            remote=False,
            random_state=42,
        )

        probe.fit(["positive"], ["negative"])
        accuracy = probe.score(["test one", "test two"], [1, 0])

        assert isinstance(accuracy, float)
        assert 0.0 <= accuracy <= 1.0

    def test_ridge_predict_proba_raises(self, tiny_model):
        """predict_proba() raises error with RidgeClassifier."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            classifier="ridge",
            device="cpu",
            remote=False,
            random_state=42,
        )

        probe.fit(["positive"], ["negative"])

        with pytest.raises(AttributeError):
            probe.predict_proba(["test input"])

    def test_ridge_multiple_predictions(self, tiny_model):
        """predict() handles multiple samples with RidgeClassifier."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            classifier="ridge",
            device="cpu",
            remote=False,
            random_state=42,
        )

        probe.fit(
            ["positive one", "positive two", "positive three"],
            ["negative one", "negative two", "negative three"],
        )
        predictions = probe.predict(["test one", "test two", "test three"])

        assert predictions.shape == (3,)
        assert all(p in [0, 1] for p in predictions)

    def test_ridge_save_load(self, tiny_model, tmp_path):
        """save/load works with RidgeClassifier."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            classifier="ridge",
            device="cpu",
            remote=False,
            random_state=42,
        )

        probe.fit(["positive"], ["negative"])
        original_pred = probe.predict(["test"])

        save_path = tmp_path / "ridge_probe.pkl"
        probe.save(str(save_path))

        loaded = LinearProbe.load(str(save_path))
        loaded_pred = loaded.predict(["test"])

        assert np.array_equal(original_pred, loaded_pred)


class TestCustomClassifierWithoutProba:
    """Tests for custom classifiers without predict_proba."""

    def test_custom_classifier_without_proba(self, tiny_model):
        """Custom classifier without predict_proba works for predict()."""
        from sklearn.linear_model import Perceptron

        # Perceptron doesn't have predict_proba by default
        clf = Perceptron(random_state=42)

        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            classifier=clf,
            device="cpu",
            remote=False,
        )

        probe.fit(["positive"], ["negative"])
        predictions = probe.predict(["test input"])

        assert predictions.shape == (1,)
        assert predictions[0] in [0, 1]

    def test_custom_classifier_score_works(self, tiny_model):
        """score() works with custom classifier without predict_proba."""
        from sklearn.linear_model import Perceptron

        clf = Perceptron(random_state=42)

        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            classifier=clf,
            device="cpu",
            remote=False,
        )

        probe.fit(["positive"], ["negative"])
        accuracy = probe.score(["test"], [1])

        assert isinstance(accuracy, float)
        assert 0.0 <= accuracy <= 1.0


class TestPerTokenPredictWithoutProba:
    """Tests for per-token prediction with classifiers lacking predict_proba."""

    def test_ridge_with_inference_pooling_all(self, tiny_model):
        """predict() with inference_pooling='all' works for ridge."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            classifier="ridge",
            pooling="last_token",
            inference_pooling="all",
            device="cpu",
            remote=False,
            random_state=42,
        )

        probe.fit(["positive example"], ["negative example"])

        # Should work and return aggregated predictions
        predictions = probe.predict(["test with multiple tokens"])

        assert predictions.shape == (1,)
        assert predictions[0] in [0, 1]

    def test_ridge_with_different_pooling(self, tiny_model):
        """predict() works with different pooling strategies for ridge."""
        for pooling in ["last_token", "mean"]:
            probe = LinearProbe(
                model=tiny_model,
                layers=-1,
                classifier="ridge",
                pooling=pooling,
                device="cpu",
                remote=False,
                random_state=42,
            )

            probe.fit(["positive"], ["negative"])
            predictions = probe.predict(["test"])

            assert predictions.shape == (1,)
            assert predictions[0] in [0, 1]


class TestAllClassifiersPredict:
    """Parametrized tests to ensure all classifiers work with predict()."""

    @pytest.mark.parametrize("classifier", list(CLASSIFICATION_CLASSIFIERS))
    def test_predict_works_for_all_builtin_classifiers(self, tiny_model, classifier):
        """Every builtin classifier works with predict()."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            classifier=classifier,
            device="cpu",
            remote=False,
            random_state=42,
        )

        # Some classifiers need more samples
        # LDA needs more samples than classes
        # LogisticRegressionCV uses 5-fold CV so needs at least 5 samples per class
        if classifier in ("lda", "logistic_regression_cv"):
            pos = [f"positive {i}" for i in range(5)]
            neg = [f"negative {i}" for i in range(5)]
        else:
            pos = ["positive"]
            neg = ["negative"]

        probe.fit(pos, neg)
        predictions = probe.predict(["test"])

        assert predictions.shape == (1,)
        assert predictions[0] in [0, 1]

    @pytest.mark.parametrize("classifier", list(CLASSIFICATION_CLASSIFIERS))
    def test_score_works_for_all_builtin_classifiers(self, tiny_model, classifier):
        """Every builtin classifier works with score()."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            classifier=classifier,
            device="cpu",
            remote=False,
            random_state=42,
        )

        # Some classifiers need more samples
        # LDA needs more samples than classes
        # LogisticRegressionCV uses 5-fold CV so needs at least 5 samples per class
        if classifier in ("lda", "logistic_regression_cv"):
            pos = [f"positive {i}" for i in range(5)]
            neg = [f"negative {i}" for i in range(5)]
        else:
            pos = ["positive"]
            neg = ["negative"]

        probe.fit(pos, neg)
        accuracy = probe.score(["test"], [1])

        assert isinstance(accuracy, float)
        assert 0.0 <= accuracy <= 1.0



"""Baseline classifiers for comparison with linear probes.

These baselines help determine whether a probe is learning something
beyond simple lexical features. If a probe doesn't beat bag-of-words,
it's likely just doing token matching.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from scipy.sparse import issparse
from sklearn.base import clone
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.dummy import DummyClassifier
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer

from .classifiers import resolve_classifier

if TYPE_CHECKING:
    from sklearn.base import BaseEstimator


class BaselineProbe:
    """Text classification baseline for comparison with linear probes.

    This class provides simple baselines that don't use model activations,
    helping determine if probes are learning meaningful representations
    or just exploiting surface-level features.

    Parameters
    ----------
    method : str, default="tfidf"
        Feature extraction method:
        - "bow": Bag-of-words (word counts)
        - "tfidf": TF-IDF weighted bag-of-words
        - "random": Random predictions (true chance baseline)
        - "majority": Always predict majority class
        - "sentence_length": Use character/word count features
        - "perplexity": Use model's own logprobs (requires model param)
        - "sentence_transformers": Use off-the-shelf embeddings
        - "shuffled_labels": Sanity check — trains on shuffled labels using
          features from base_method. Should get ~50% accuracy.
    classifier : str | BaseEstimator, default="logistic_regression"
        Classification model. Same options as LinearProbe.
        Ignored for method="random" and method="majority".
    random_state : int | None, default=None
        Random seed for reproducibility.
    max_features : int | None, default=10000
        Maximum vocabulary size for bow/tfidf methods.
    ngram_range : tuple[int, int], default=(1, 1)
        N-gram range for bow/tfidf. (1, 1) = unigrams only,
        (1, 2) = unigrams and bigrams.
    base_method : str, default="tfidf"
        Feature extraction method to use when method="shuffled_labels".
        Ignored for other methods.
    model : str | None, default=None
        HuggingFace model ID. Required for method="perplexity".
    device : str, default="auto"
        Device for model inference (for perplexity method).
    remote : bool, default=False
        Use nnsight remote execution (for perplexity method).

    Attributes
    ----------
    classifier_ : BaseEstimator
        The fitted classifier (after calling fit()).
    classes_ : np.ndarray
        Class labels (after calling fit()).
    vectorizer_ : CountVectorizer | TfidfVectorizer | None
        The fitted text vectorizer (for bow/tfidf methods).

    Examples
    --------
    >>> baseline = BaselineProbe(method="tfidf", classifier="logistic_regression")
    >>> baseline.fit(positive_prompts, negative_prompts)
    >>> accuracy = baseline.score(test_prompts, test_labels)
    >>> print(f"TF-IDF baseline: {accuracy:.1%}")
    """

    VALID_METHODS = (
        "bow",
        "tfidf",
        "random",
        "majority",
        "sentence_length",
        "perplexity",
        "sentence_transformers",
        "shuffled_labels",
    )

    def __init__(
        self,
        method: str = "tfidf",
        classifier: str | BaseEstimator = "logistic_regression",
        random_state: int | None = None,
        max_features: int | None = 10000,
        ngram_range: tuple[int, int] = (1, 1),
        base_method: str = "tfidf",
        model: str | None = None,
        device: str = "auto",
        remote: bool = False,
    ):
        if method not in self.VALID_METHODS:
            raise ValueError(
                f"Unknown method: {method!r}. Valid options: {self.VALID_METHODS}"
            )

        if method == "perplexity" and model is None:
            raise ValueError(
                "method='perplexity' requires the 'model' parameter to be specified."
            )

        if method == "shuffled_labels" and base_method not in self.VALID_METHODS:
            raise ValueError(
                f"Unknown base_method: {base_method!r}. "
                f"Valid options: {self.VALID_METHODS}"
            )

        self.method = method
        self.classifier = classifier
        self.random_state = random_state
        self.max_features = max_features
        self.ngram_range = ngram_range
        self.base_method = base_method
        self.model = model
        self.device = device
        self.remote = remote

        # Resolve classifier template
        if method in ("random", "majority"):
            strategy = "uniform" if method == "random" else "most_frequent"
            self._classifier_template = DummyClassifier(
                strategy=strategy, random_state=random_state
            )
        else:
            self._classifier_template = resolve_classifier(classifier, random_state)

        # Fitted state
        self.classifier_: BaseEstimator | None = None
        self.classes_: np.ndarray | None = None
        self.vectorizer_: CountVectorizer | TfidfVectorizer | None = None
        self._st_model = None  # Sentence transformer model (lazy loaded)

    def fit(
        self,
        positive_prompts: list[str],
        negative_prompts: list[str],
    ) -> BaselineProbe:
        """Fit the baseline on contrastive examples.

        Parameters
        ----------
        positive_prompts : list[str]
            Examples of the positive class.
        negative_prompts : list[str]
            Examples of the negative class.

        Returns
        -------
        BaselineProbe
            Self, for method chaining.
        """
        # Combine prompts and create labels
        prompts = positive_prompts + negative_prompts
        labels = np.array([1] * len(positive_prompts) + [0] * len(negative_prompts))

        # For shuffled_labels, use base_method for features
        feature_method = self.base_method if self.method == "shuffled_labels" else self.method

        # Extract features based on method
        if feature_method == "bow":
            self.vectorizer_ = CountVectorizer(
                max_features=self.max_features,
                ngram_range=self.ngram_range,
            )
            X = self.vectorizer_.fit_transform(prompts)
        elif feature_method == "tfidf":
            self.vectorizer_ = TfidfVectorizer(
                max_features=self.max_features,
                ngram_range=self.ngram_range,
            )
            X = self.vectorizer_.fit_transform(prompts)
        elif feature_method == "sentence_length":
            X = self._extract_sentence_length_features(prompts)
        elif feature_method == "perplexity":
            X = self._compute_perplexity(prompts)
        elif feature_method == "sentence_transformers":
            X = self._transform_sentence_transformers(prompts, fit=True)
        else:
            # random/majority - features don't matter, just need shape
            X = np.zeros((len(prompts), 1))

        # Shuffle labels for sanity check baseline
        if self.method == "shuffled_labels":
            rng = np.random.RandomState(self.random_state)
            labels = rng.permutation(labels)

        # Fit classifier
        self.classifier_ = clone(self._classifier_template)
        # Some classifiers (LDA) don't support sparse matrices
        X = self._ensure_dense_if_needed(X)
        self.classifier_.fit(X, labels)
        self.classes_ = self.classifier_.classes_

        return self

    def predict(self, prompts: list[str]) -> np.ndarray:
        """Predict class labels for prompts.

        Parameters
        ----------
        prompts : list[str]
            Text prompts to classify.

        Returns
        -------
        np.ndarray
            Predicted class labels, shape (n_prompts,).
        """
        self._check_fitted()
        X = self._transform(prompts)
        return self.classifier_.predict(X)

    def predict_proba(self, prompts: list[str]) -> np.ndarray:
        """Predict class probabilities for prompts.

        Parameters
        ----------
        prompts : list[str]
            Text prompts to classify.

        Returns
        -------
        np.ndarray
            Class probabilities, shape (n_prompts, n_classes).

        Raises
        ------
        AttributeError
            If the classifier doesn't support predict_proba.
        """
        self._check_fitted()

        if not hasattr(self.classifier_, "predict_proba"):
            raise AttributeError(
                f"Classifier {type(self.classifier_).__name__} does not support "
                "predict_proba(). Use a classifier like logistic_regression or lda."
            )

        X = self._transform(prompts)
        return self.classifier_.predict_proba(X)

    def score(self, prompts: list[str], labels: list[int] | np.ndarray) -> float:
        """Compute classification accuracy.

        Parameters
        ----------
        prompts : list[str]
            Text prompts to classify.
        labels : list[int] | np.ndarray
            True labels.

        Returns
        -------
        float
            Classification accuracy.
        """
        self._check_fitted()
        predictions = self.predict(prompts)
        labels = np.asarray(labels)
        return float(np.mean(predictions == labels))

    def _transform(self, prompts: list[str]) -> np.ndarray:
        """Transform prompts to feature matrix."""
        feature_method = self.base_method if self.method == "shuffled_labels" else self.method

        if feature_method in ("bow", "tfidf"):
            X = self.vectorizer_.transform(prompts)
            return self._ensure_dense_if_needed(X)
        elif feature_method == "sentence_length":
            return self._extract_sentence_length_features(prompts)
        elif feature_method == "perplexity":
            return self._compute_perplexity(prompts)
        elif feature_method == "sentence_transformers":
            return self._transform_sentence_transformers(prompts, fit=False)
        else:
            # random/majority
            return np.zeros((len(prompts), 1))

    def _ensure_dense_if_needed(self, X: np.ndarray) -> np.ndarray:
        """Convert sparse matrix to dense if classifier requires it."""
        if issparse(X) and isinstance(self._classifier_template, LinearDiscriminantAnalysis):
            return X.toarray()
        return X

    def _extract_sentence_length_features(self, prompts: list[str]) -> np.ndarray:
        """Extract sentence length features from prompts.

        Features: character count, word count, average word length.
        """
        features = []
        for prompt in prompts:
            char_count = len(prompt)
            word_count = len(prompt.split())
            avg_word_len = char_count / max(word_count, 1)
            features.append([char_count, word_count, avg_word_len])
        return np.array(features)

    def _compute_perplexity(self, prompts: list[str]) -> np.ndarray:
        """Compute perplexity features for each prompt.

        Uses the model's own logprobs to compute perplexity.
        Features: mean perplexity, min perplexity, max perplexity.

        Results are cached per-prompt to disk to avoid redundant NDIF calls
        and enable cross-run reuse.
        """
        import torch
        from tqdm import tqdm

        from .cache import (
            is_prompt_perplexity_cached,
            load_prompt_perplexity,
            save_prompt_perplexity,
        )
        from .extraction import _require_nnsight, configure_remote, get_cached_model

        _require_nnsight()

        # Check which prompts are already cached
        cached_indices = []
        uncached_prompts = []
        uncached_indices = []

        for i, prompt in enumerate(prompts):
            if is_prompt_perplexity_cached(self.model, prompt):
                cached_indices.append(i)
            else:
                uncached_prompts.append(prompt)
                uncached_indices.append(i)

        # Initialize result array
        features = np.zeros((len(prompts), 3), dtype=np.float32)

        # Load cached prompts
        for i in cached_indices:
            ppl = load_prompt_perplexity(self.model, prompts[i])
            features[i] = ppl.float().numpy()

        # If all prompts are cached, return early
        if not uncached_prompts:
            return features

        # Configure remote if needed
        if self.remote:
            configure_remote()

        model = get_cached_model(self.model, self.device, remote=self.remote)

        # Compute perplexity for uncached prompts
        for idx, prompt in tqdm(
            zip(uncached_indices, uncached_prompts),
            total=len(uncached_prompts),
            desc="Computing perplexity",
            unit="prompt",
        ):
            # Tokenize
            inputs = model.tokenizer(prompt, return_tensors="pt")

            # Forward pass to get logits using nnsight tracing
            with model.trace(inputs, remote=self.remote):
                logits = model.lm_head.output.save()

            # Unwrap proxy if needed
            logits_val = logits.value if hasattr(logits, "value") else logits

            # Compute per-token cross-entropy loss
            # Shift logits and labels for next-token prediction
            shift_logits = logits_val[..., :-1, :].contiguous()
            shift_labels = inputs["input_ids"][..., 1:].contiguous()

            # Move to same device
            if shift_logits.device != shift_labels.device:
                shift_labels = shift_labels.to(shift_logits.device)

            loss_fn = torch.nn.CrossEntropyLoss(reduction="none")
            per_token_loss = loss_fn(
                shift_logits.view(-1, shift_logits.size(-1)),
                shift_labels.view(-1),
            )

            # Compute perplexity statistics
            mean_loss = per_token_loss.mean().item()
            min_loss = per_token_loss.min().item()
            max_loss = per_token_loss.max().item()

            mean_ppl = float(np.exp(mean_loss))
            min_ppl = float(np.exp(min_loss))
            max_ppl = float(np.exp(max_loss))

            # Save to per-prompt cache
            ppl_tensor = torch.tensor([mean_ppl, min_ppl, max_ppl], dtype=torch.float32)
            save_prompt_perplexity(self.model, prompt, ppl_tensor)

            features[idx] = ppl_tensor.float().numpy()

        return features

    def _check_sentence_transformers_installed(self) -> None:
        """Check that sentence-transformers is installed."""
        try:
            import sentence_transformers  # noqa: F401
        except ImportError:
            raise ImportError(
                "sentence-transformers is required for method='sentence_transformers'. "
                "Install it with: pip install lmprobe[embeddings]"
            )

    def _transform_sentence_transformers(
        self, prompts: list[str], fit: bool = False
    ) -> np.ndarray:
        """Extract sentence-transformer embeddings.

        Parameters
        ----------
        prompts : list[str]
            Text prompts to embed.
        fit : bool
            If True, load the sentence transformer model.

        Returns
        -------
        np.ndarray
            Embeddings with shape (n_prompts, embedding_dim).
        """
        self._check_sentence_transformers_installed()
        from sentence_transformers import SentenceTransformer

        if fit or self._st_model is None:
            # Use a small, fast model by default
            self._st_model = SentenceTransformer("all-MiniLM-L6-v2")

        embeddings = self._st_model.encode(prompts, show_progress_bar=False)
        return embeddings

    def _check_fitted(self) -> None:
        """Check if the baseline has been fitted."""
        if self.classifier_ is None:
            raise RuntimeError(
                "BaselineProbe has not been fitted. Call fit() first."
            )

    def get_feature_names(self) -> list[str] | None:
        """Get feature names for bow/tfidf methods.

        Returns
        -------
        list[str] | None
            Feature names (vocabulary), or None for random/majority.
        """
        if self.vectorizer_ is None:
            return None
        return self.vectorizer_.get_feature_names_out().tolist()

    def get_top_features(self, n: int = 20) -> dict[str, list[tuple[str, float]]] | None:
        """Get top features by classifier weight for each class.

        Only works for bow/tfidf with linear classifiers that have coef_.

        Parameters
        ----------
        n : int, default=20
            Number of top features to return per class.

        Returns
        -------
        dict | None
            Dictionary with 'positive' and 'negative' keys, each containing
            a list of (feature_name, weight) tuples. None if not applicable.
        """
        if self.vectorizer_ is None or self.classifier_ is None:
            return None

        if not hasattr(self.classifier_, "coef_"):
            return None

        feature_names = self.get_feature_names()
        coef = self.classifier_.coef_.ravel()

        # Get indices sorted by weight
        top_positive_idx = np.argsort(coef)[-n:][::-1]
        top_negative_idx = np.argsort(coef)[:n]

        return {
            "positive": [(feature_names[i], coef[i]) for i in top_positive_idx],
            "negative": [(feature_names[i], coef[i]) for i in top_negative_idx],
        }



"""Per-layer feature scaling for multi-layer probes.

When using activations from multiple layers, each layer may have different
activation magnitude distributions. This module provides scalers that normalize
each layer's features independently to enable fair comparison.
"""

from __future__ import annotations

import numpy as np


class PerLayerScaler:
    """Standardize features on a per-layer basis.

    When using multiple layers (concatenated), each layer may have different
    activation magnitude distributions. This scaler normalizes each layer's
    features to zero mean and unit variance.

    Two strategies are available:
    - "per_neuron": Each neuron gets its own mean/std (more parameters, higher variance)
    - "per_layer": All neurons in a layer share one mean/std (fewer parameters, lower variance)

    The "per_layer" strategy may be preferable when:
    - Sample size is small relative to hidden dimension
    - Neurons within a layer have similar activation distributions (symmetry assumption)

    Parameters
    ----------
    n_layers : int
        Number of layers in the concatenated features.
    hidden_dim : int
        Hidden dimension per layer (features per layer).
    strategy : str, default="per_neuron"
        Scaling strategy:
        - "per_neuron": Each neuron has its own mean/std
        - "per_layer": All neurons in a layer share one mean/std

    Attributes
    ----------
    means_ : np.ndarray | None
        Feature means. Shape depends on strategy:
        - "per_neuron": (n_layers, hidden_dim)
        - "per_layer": (n_layers,)
    stds_ : np.ndarray | None
        Feature standard deviations. Shape matches means_.

    Examples
    --------
    >>> scaler = PerLayerScaler(n_layers=3, hidden_dim=128, strategy="per_neuron")
    >>> X_train_scaled = scaler.fit_transform(X_train)
    >>> X_test_scaled = scaler.transform(X_test)

    >>> # Use per_layer for small sample sizes
    >>> scaler = PerLayerScaler(n_layers=3, hidden_dim=128, strategy="per_layer")
    """

    VALID_STRATEGIES = ("per_neuron", "per_layer")

    def __init__(
        self, n_layers: int, hidden_dim: int, strategy: str = "per_neuron"
    ):
        if strategy not in self.VALID_STRATEGIES:
            raise ValueError(
                f"Unknown strategy: {strategy!r}. "
                f"Valid options: {self.VALID_STRATEGIES}"
            )
        self.n_layers = n_layers
        self.hidden_dim = hidden_dim
        self.strategy = strategy
        self.means_: np.ndarray | None = None
        self.stds_: np.ndarray | None = None

    def fit(self, X: np.ndarray) -> PerLayerScaler:
        """Compute per-layer means and standard deviations.

        Parameters
        ----------
        X : np.ndarray
            Feature matrix, shape (n_samples, n_layers * hidden_dim).

        Returns
        -------
        PerLayerScaler
            Self, for method chaining.

        Raises
        ------
        ValueError
            If X has wrong number of features.
        """
        expected_features = self.n_layers * self.hidden_dim
        if X.shape[1] != expected_features:
            raise ValueError(
                f"Expected {expected_features} features "
                f"({self.n_layers} layers x {self.hidden_dim} hidden_dim), "
                f"got {X.shape[1]}"
            )

        # Reshape to (n_samples, n_layers, hidden_dim)
        X_reshaped = X.reshape(X.shape[0], self.n_layers, self.hidden_dim)

        if self.strategy == "per_neuron":
            # Each neuron gets its own mean/std
            # Mean over samples, shape: (n_layers, hidden_dim)
            self.means_ = X_reshaped.mean(axis=0)
            self.stds_ = X_reshaped.std(axis=0)
        else:  # per_layer
            # All neurons in a layer share one mean/std
            # Reshape to (n_layers, n_samples * hidden_dim) and compute stats
            X_flat = X_reshaped.transpose(1, 0, 2).reshape(self.n_layers, -1)
            self.means_ = X_flat.mean(axis=1)  # shape: (n_layers,)
            self.stds_ = X_flat.std(axis=1)  # shape: (n_layers,)

        # Avoid division by zero: replace zero std with 1
        self.stds_ = np.where(self.stds_ == 0, 1.0, self.stds_)

        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        """Apply per-layer standardization.

        Parameters
        ----------
        X : np.ndarray
            Feature matrix, shape (n_samples, n_layers * hidden_dim).

        Returns
        -------
        np.ndarray
            Standardized features, same shape as input.

        Raises
        ------
        RuntimeError
            If scaler has not been fitted.
        ValueError
            If X has wrong number of features.
        """
        if self.means_ is None or self.stds_ is None:
            raise RuntimeError("PerLayerScaler has not been fitted. Call fit() first.")

        expected_features = self.n_layers * self.hidden_dim
        if X.shape[1] != expected_features:
            raise ValueError(
                f"Expected {expected_features} features "
                f"({self.n_layers} layers x {self.hidden_dim} hidden_dim), "
                f"got {X.shape[1]}"
            )

        # Reshape to (n_samples, n_layers, hidden_dim)
        X_reshaped = X.reshape(X.shape[0], self.n_layers, self.hidden_dim)

        if self.strategy == "per_neuron":
            # means_ and stds_ are (n_layers, hidden_dim)
            X_scaled = (X_reshaped - self.means_) / self.stds_
        else:  # per_layer
            # means_ and stds_ are (n_layers,), need to broadcast
            # Reshape to (1, n_layers, 1) for broadcasting
            means = self.means_[:, np.newaxis]  # (n_layers, 1)
            stds = self.stds_[:, np.newaxis]  # (n_layers, 1)
            X_scaled = (X_reshaped - means) / stds

        # Reshape back to (n_samples, n_layers * hidden_dim)
        return X_scaled.reshape(X.shape[0], -1)

    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        """Fit scaler and transform data in one step.

        Parameters
        ----------
        X : np.ndarray
            Feature matrix, shape (n_samples, n_layers * hidden_dim).

        Returns
        -------
        np.ndarray
            Standardized features, same shape as input.
        """
        return self.fit(X).transform(X)

    def inverse_transform(self, X: np.ndarray) -> np.ndarray:
        """Reverse the standardization.

        Parameters
        ----------
        X : np.ndarray
            Standardized feature matrix, shape (n_samples, n_layers * hidden_dim).

        Returns
        -------
        np.ndarray
            Original-scale features, same shape as input.

        Raises
        ------
        RuntimeError
            If scaler has not been fitted.
        """
        if self.means_ is None or self.stds_ is None:
            raise RuntimeError("PerLayerScaler has not been fitted. Call fit() first.")

        expected_features = self.n_layers * self.hidden_dim
        if X.shape[1] != expected_features:
            raise ValueError(
                f"Expected {expected_features} features, got {X.shape[1]}"
            )

        # Reshape to (n_samples, n_layers, hidden_dim)
        X_reshaped = X.reshape(X.shape[0], self.n_layers, self.hidden_dim)

        if self.strategy == "per_neuron":
            # means_ and stds_ are (n_layers, hidden_dim)
            X_original = X_reshaped * self.stds_ + self.means_
        else:  # per_layer
            # means_ and stds_ are (n_layers,), need to broadcast
            means = self.means_[:, np.newaxis]  # (n_layers, 1)
            stds = self.stds_[:, np.newaxis]  # (n_layers, 1)
            X_original = X_reshaped * stds + means

        # Reshape back
        return X_original.reshape(X.shape[0], -1)

    def get_layer_stats(self) -> dict[str, np.ndarray]:
        """Get per-layer statistics for analysis.

        Returns
        -------
        dict
            Dictionary with layer-level statistics. Contents depend on strategy:
            - "per_neuron": includes 'mean_norms', 'std_norms', 'mean_per_layer', 'std_per_layer'
            - "per_layer": includes 'means', 'stds' (the raw per-layer values)

        Raises
        ------
        RuntimeError
            If scaler has not been fitted.
        """
        if self.means_ is None or self.stds_ is None:
            raise RuntimeError("PerLayerScaler has not been fitted. Call fit() first.")

        if self.strategy == "per_neuron":
            return {
                "mean_norms": np.linalg.norm(self.means_, axis=1),
                "std_norms": np.linalg.norm(self.stds_, axis=1),
                "mean_per_layer": self.means_.mean(axis=1),
                "std_per_layer": self.stds_.mean(axis=1),
            }
        else:  # per_layer
            return {
                "means": self.means_.copy(),
                "stds": self.stds_.copy(),
            }



"""Unified cache for extracting activations and perplexity in a single forward pass.

This module provides the UnifiedCache class which optimizes extraction by capturing
both layer activations and logits (for perplexity) in a single nnsight trace.
"""

from __future__ import annotations

import logging
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np
import torch
from tqdm import tqdm

from .backends import resolve_backend
from .cache import (
    get_prompt_cached_pooled_layers,
    get_prompt_cached_raw_layers,
    is_prompt_fully_cached,
    is_prompt_logits_cached,
    is_prompt_perplexity_cached,
    is_prompt_pooled_cached,
    load_prompt_activations,
    load_prompt_logits,
    load_prompt_perplexity,
    load_prompt_pooled_activations,
    save_prompt_activations,
    save_prompt_logits,
    save_prompt_perplexity,
    save_prompt_pooled_activations,
)
from .extraction import (
    compute_perplexity_from_logits,
    get_num_layers_from_config,
    resolve_layers,
)
from .pooling import TRAIN_POOLING_STRATEGIES, get_pooling_fn

if TYPE_CHECKING:
    from .backends import ExtractionBackend


logger = logging.getLogger(__name__)


@dataclass
class WarmupStats:
    """Statistics from a warmup operation."""

    total_prompts: int
    activations_cached: int
    activations_extracted: int
    perplexity_cached: int
    perplexity_extracted: int
    elapsed_seconds: float
    logits_cached: int = 0
    logits_extracted: int = 0

    @property
    def cache_hit_rate(self) -> float:
        """Fraction of prompts that had activations cached."""
        if self.total_prompts == 0:
            return 0.0
        return self.activations_cached / self.total_prompts

    def __repr__(self) -> str:
        parts = [
            f"WarmupStats(total={self.total_prompts}",
            f"activations={self.activations_cached} cached "
            f"+ {self.activations_extracted} extracted",
            f"perplexity={self.perplexity_cached} cached "
            f"+ {self.perplexity_extracted} extracted",
        ]
        if self.logits_cached or self.logits_extracted:
            parts.append(
                f"logits={self.logits_cached} cached "
                f"+ {self.logits_extracted} extracted"
            )
        parts.append(f"time={self.elapsed_seconds:.1f}s)")
        return ", ".join(parts)


@dataclass
class CachedLogits:
    """Cached logits from language model extraction.

    Attributes
    ----------
    values : torch.Tensor
        Logit values. Shape (batch, positions, vocab_size) for full logits,
        or (batch, positions, K) for top-k logits.
    indices : torch.Tensor | None
        Top-k token indices with shape (batch, positions, K) and dtype int32.
        None when full logits are stored.
    top_k : int | None
        K value if top-k logits, None for full logits.
    positions : str
        Which token positions are stored: "last" or "all".
    """

    values: torch.Tensor
    indices: torch.Tensor | None
    top_k: int | None
    positions: str


class UnifiedCache:
    """Extracts activations and perplexity in a single forward pass.

    This class provides efficient extraction by capturing both layer activations
    and lm_head logits in a single nnsight trace, then computing perplexity
    features from the logits.

    Parameters
    ----------
    model : str
        HuggingFace model ID or local path.
    layers : int | list[int] | str, default="all"
        Which layers to extract activations from:
        - int: Single layer (negative indexing supported)
        - list[int]: Multiple layers
        - "all": All layers (default)
        - "middle": Middle third of layers
        - "last": Last layer only
    compute_perplexity : bool, default=True
        Whether to also capture logits and compute perplexity features.
    device : str, default="auto"
        Device for model inference.
    remote : bool, default=False
        Use nnsight remote execution (requires NNSIGHT_API_KEY).
    batch_size : int, default=8
        Number of prompts to process at once.
    cache_pooled : bool, default=True
        If True (default), apply pooling BEFORE caching and store only
        the pooled activations. This reduces disk usage by ~100x (storing
        only (1, hidden_dim) per layer instead of (1, seq_len, hidden_dim)).

        Set to False only if you need to re-pool cached activations with
        different strategies.

        WARNING: When cache_pooled=True, you must use the same pooling
        strategy for all operations. The cached data cannot be re-pooled
        with a different strategy.
    pooling : str, default="last_token"
        Pooling strategy to use when cache_pooled=True. Options:
        - "last_token": Use the last non-padding token (default)
        - "first_token": Use the first token
        - "mean": Mean of all non-padding tokens
    backend : str, default="local"
        Extraction backend: "local" (default) or "nnsight".
    dtype : str or None, default=None
        Model dtype as a string: "float32", "float16", or "bfloat16".
        If None, defaults to float32 for local backend.

    Examples
    --------
    >>> # Memory-efficient caching (recommended for large models)
    >>> cache = UnifiedCache(
    ...     model="meta-llama/Llama-3.1-70B",
    ...     layers="all",
    ...     cache_pooled=True,      # Store only pooled activations
    ...     pooling="last_token",   # ~100x less disk space
    ... )
    >>> stats = cache.warmup(prompts)
    >>> print(f"Extracted {stats.activations_extracted} prompts")
    """

    def __init__(
        self,
        model: str,
        layers: int | list[int] | str = "all",
        compute_perplexity: bool = True,
        device: str = "auto",
        remote: bool = False,
        batch_size: int = 8,
        cache_pooled: bool = True,
        pooling: str = "last_token",
        backend: str = "local",
        dtype: str | None = None,
        cache_logits: bool = False,
        logit_top_k: int | None = None,
        logit_positions: str = "last",
    ):
        self.model_name = model
        self.layers_spec = layers
        self.compute_perplexity = compute_perplexity
        self.device = device
        self.remote = remote
        self.batch_size = batch_size
        self.cache_pooled = cache_pooled
        self.pooling = pooling
        self.backend_name = backend
        self.dtype = dtype
        self.cache_logits = cache_logits
        self.logit_top_k = logit_top_k
        self.logit_positions = logit_positions

        # Validate pooling strategy
        if cache_pooled and pooling not in TRAIN_POOLING_STRATEGIES:
            raise ValueError(
                f"Invalid pooling strategy for cache_pooled: {pooling!r}. "
                f"Available: {sorted(TRAIN_POOLING_STRATEGIES - {'all'})}"
            )
        if cache_pooled and pooling == "all":
            raise ValueError(
                "pooling='all' is not valid with cache_pooled=True. "
                "Use 'last_token', 'first_token', or 'mean'."
            )

        if logit_positions not in ("last", "all"):
            raise ValueError(
                f"Invalid logit_positions: {logit_positions!r}. "
                f"Must be 'last' or 'all'."
            )

        # Create backend (lazy-loads model)
        self._backend: ExtractionBackend | None = None
        self._layer_indices: list[int] | None = None
        self._pooling_fn = get_pooling_fn(pooling) if cache_pooled else None

    @property
    def _resolved_backend(self) -> ExtractionBackend:
        """Get the backend, creating if necessary."""
        if self._backend is None:
            torch_dtype = self._resolve_dtype(self.dtype)
            self._backend = resolve_backend(
                self.backend_name, self.model_name, self.device,
                remote=self.remote, dtype=torch_dtype,
            )
        return self._backend

    @staticmethod
    def _resolve_dtype(dtype: str | None) -> torch.dtype | None:
        """Resolve a dtype string to a torch.dtype, or None."""
        if dtype is None:
            return None
        _dtype_map = {
            "float32": torch.float32,
            "float16": torch.float16,
            "bfloat16": torch.bfloat16,
        }
        if dtype not in _dtype_map:
            raise ValueError(
                f"Unknown dtype: {dtype!r}. "
                f"Available: {list(_dtype_map.keys())}"
            )
        return _dtype_map[dtype]

    @property
    def model(self):
        """Get the loaded model, loading if necessary."""
        return self._resolved_backend.model

    @property
    def layer_indices(self) -> list[int]:
        """Get resolved layer indices."""
        if self._layer_indices is None:
            num_layers = get_num_layers_from_config(self.model_name)
            self._layer_indices = resolve_layers(self.layers_spec, num_layers)
        return self._layer_indices

    def _check_cache_status(
        self, prompts: list[str]
    ) -> tuple[list[str], list[str], list[str]]:
        """Check which prompts need extraction.

        Returns
        -------
        tuple
            - prompts needing activation extraction
            - prompts needing perplexity extraction (if compute_perplexity=True)
            - prompts needing logit extraction (if cache_logits=True)
        """
        required_layers = set(self.layer_indices)

        need_activations = []
        need_perplexity = []
        need_logits = []
        partial_cache_count = 0
        partial_cache_found_layers: set[int] | None = None

        for prompt in prompts:
            # Check appropriate cache based on cache_pooled setting
            if self.cache_pooled:
                act_cached = is_prompt_pooled_cached(
                    self.model_name, prompt, required_layers, self.pooling
                )
            else:
                act_cached = is_prompt_fully_cached(
                    self.model_name, prompt, required_layers
                )

            ppl_cached = (
                is_prompt_perplexity_cached(self.model_name, prompt)
                if self.compute_perplexity
                else True
            )

            logits_cached = (
                is_prompt_logits_cached(
                    self.model_name, prompt, self.logit_top_k
                )
                if self.cache_logits
                else True
            )

            if not act_cached:
                need_activations.append(prompt)
                # Check if cache file exists but is missing requested layers
                if self.cache_pooled:
                    cached_layers = get_prompt_cached_pooled_layers(
                        self.model_name, prompt, self.pooling
                    )
                else:
                    cached_layers = get_prompt_cached_raw_layers(
                        self.model_name, prompt
                    )
                if cached_layers is not None and len(cached_layers) > 0:
                    partial_cache_count += 1
                    if partial_cache_found_layers is None:
                        partial_cache_found_layers = cached_layers

            if not ppl_cached:
                need_perplexity.append(prompt)

            if not logits_cached:
                need_logits.append(prompt)

        if partial_cache_count > 0:
            missing = sorted(required_layers - (partial_cache_found_layers or set()))
            found = sorted(partial_cache_found_layers or set())
            import warnings

            warnings.warn(
                f"Cache exists for {partial_cache_count} prompt(s) but missing "
                f"layer(s) {missing} (found layers {found}). "
                f"A new forward pass is required. To avoid redundant forward passes, "
                f"use probe.warmup(prompts) to cache all layers at once before fitting.",
                stacklevel=2,
            )

        return need_activations, need_perplexity, need_logits

    def warmup(
        self,
        prompts: list[str],
        remote: bool | None = None,
    ) -> WarmupStats:
        """Extract and cache activations/perplexity for prompts.

        This method checks the cache, identifies what needs extraction,
        and performs minimal forward passes to fill the cache.

        Parameters
        ----------
        prompts : list[str]
            Text prompts to warm up the cache for.
        remote : bool | None
            Override the instance-level remote setting.

        Returns
        -------
        WarmupStats
            Statistics about the warmup operation.
        """
        start_time = time.time()
        remote = self.remote if remote is None else remote

        if remote and self.backend_name == "nnsight":
            from .extraction import configure_remote

            configure_remote()

        layer_indices = sorted(self.layer_indices)

        # Check cache status
        need_activations, need_perplexity, need_logits = self._check_cache_status(
            prompts
        )

        # Compute which prompts need unified extraction
        # (prompts that need BOTH or where we can get both cheaply)
        need_activations_set = set(need_activations)
        need_perplexity_set = set(need_perplexity)
        need_logits_set = set(need_logits)
        need_unified = need_activations_set | need_perplexity_set | need_logits_set
        need_unified_list = [p for p in prompts if p in need_unified]

        activations_cached = len(prompts) - len(need_activations)
        perplexity_cached = len(prompts) - len(need_perplexity)
        logits_cached = len(prompts) - len(need_logits) if self.cache_logits else 0
        activations_extracted = 0
        perplexity_extracted = 0
        logits_extracted = 0

        logger.info(
            f"[UNIFIED] Checking cache for {len(prompts)} prompts, "
            f"layers: {layer_indices}"
        )
        logger.info(
            f"[UNIFIED] Activations: {activations_cached} cached, "
            f"{len(need_activations)} need extraction"
        )
        if self.compute_perplexity:
            logger.info(
                f"[UNIFIED] Perplexity: {perplexity_cached} cached, "
                f"{len(need_perplexity)} need extraction"
            )
        if self.cache_logits:
            logger.info(
                f"[UNIFIED] Logits: {logits_cached} cached, "
                f"{len(need_logits)} need extraction"
            )

        if need_unified_list:
            backend = self._resolved_backend
            num_batches = (len(need_unified_list) + self.batch_size - 1) // self.batch_size

            logger.info(
                f"[UNIFIED] Extracting {len(need_unified_list)} prompts in "
                f"{num_batches} batches"
            )

            with torch.no_grad():
                for batch_idx in tqdm(
                    range(0, len(need_unified_list), self.batch_size),
                    total=num_batches,
                    desc="Unified extraction",
                    unit="batch",
                ):
                    batch_prompts = need_unified_list[
                        batch_idx : batch_idx + self.batch_size
                    ]

                    # Single forward pass captures both activations and logits
                    batch_acts, batch_mask, batch_logits = (
                        backend.extract_batch_with_logits(
                            batch_prompts, layer_indices, remote=remote
                        )
                    )

                    # Compute perplexity features from logits
                    if self.compute_perplexity:
                        # Get input_ids for perplexity computation
                        tokenized = backend.tokenizer(
                            batch_prompts,
                            return_tensors="pt",
                            padding=True,
                        )
                        ppl_features = compute_perplexity_from_logits(
                            batch_logits,
                            tokenized["input_ids"],
                            batch_mask,
                        )

                    # Save each prompt's data
                    for j, prompt in enumerate(batch_prompts):
                        # Save activations if needed
                        if prompt in need_activations_set:
                            prompt_acts = batch_acts[j : j + 1]
                            prompt_mask = batch_mask[j : j + 1]

                            if self.cache_pooled:
                                # Pool before saving - ~100x less disk space!
                                # prompt_acts shape: (1, seq_len, n_layers * hidden_dim)
                                pooled_acts = self._pooling_fn(prompt_acts, prompt_mask)
                                # pooled_acts shape: (1, n_layers * hidden_dim)
                                save_prompt_pooled_activations(
                                    self.model_name,
                                    prompt,
                                    layer_indices,
                                    pooled_acts,
                                    self.pooling,
                                )
                            else:
                                # Save full sequence (original behavior)
                                save_prompt_activations(
                                    self.model_name,
                                    prompt,
                                    layer_indices,
                                    prompt_acts,
                                    prompt_mask,
                                )
                            activations_extracted += 1

                        # Save perplexity if needed
                        if self.compute_perplexity and prompt in need_perplexity_set:
                            save_prompt_perplexity(
                                self.model_name, prompt, ppl_features[j]
                            )
                            perplexity_extracted += 1

                        # Save logits if needed
                        if self.cache_logits and prompt in need_logits_set:
                            save_prompt_logits(
                                self.model_name,
                                prompt,
                                batch_logits[j : j + 1],
                                batch_mask[j : j + 1],
                                self.logit_top_k,
                                self.logit_positions,
                            )
                            logits_extracted += 1

        elapsed = time.time() - start_time

        stats = WarmupStats(
            total_prompts=len(prompts),
            activations_cached=activations_cached,
            activations_extracted=activations_extracted,
            perplexity_cached=perplexity_cached,
            perplexity_extracted=perplexity_extracted,
            logits_cached=logits_cached,
            logits_extracted=logits_extracted,
            elapsed_seconds=elapsed,
        )

        logger.info(
            f"[UNIFIED] Complete: {activations_extracted} activations + "
            f"{perplexity_extracted} perplexity + "
            f"{logits_extracted} logits extracted in {elapsed:.1f}s"
        )

        return stats

    def get_activations(
        self,
        prompts: list[str],
        remote: bool | None = None,
    ) -> tuple[torch.Tensor, torch.Tensor | None]:
        """Get activations for prompts (from cache or via extraction).

        Parameters
        ----------
        prompts : list[str]
            Text prompts.
        remote : bool | None
            Override the instance-level remote setting.

        Returns
        -------
        tuple[torch.Tensor, torch.Tensor | None]
            If cache_pooled=False:
                (activations, attention_mask) where activations has shape
                (batch, seq_len, n_layers * hidden_dim).
            If cache_pooled=True:
                (activations, None) where activations has shape
                (batch, n_layers * hidden_dim). Already pooled, no mask needed.
        """
        # Ensure cache is warm
        self.warmup(prompts, remote=remote)

        layer_indices = sorted(self.layer_indices)

        if self.cache_pooled:
            # Load pooled activations - already aggregated, no padding needed
            all_activations = []
            for prompt in prompts:
                acts = load_prompt_pooled_activations(
                    self.model_name, prompt, layer_indices, self.pooling
                )
                all_activations.append(acts)

            # Concatenate along batch dimension
            # Each acts has shape (1, n_layers * hidden_dim)
            return torch.cat(all_activations, dim=0), None

        else:
            # Load full-sequence activations (original behavior)
            all_activations = []
            all_masks = []

            for prompt in prompts:
                acts, mask = load_prompt_activations(
                    self.model_name, prompt, layer_indices
                )
                all_activations.append(acts)
                all_masks.append(mask)

            # Pad and concatenate
            max_seq_len = max(a.shape[1] for a in all_activations)
            hidden_dim = all_activations[0].shape[2]

            padded_acts = []
            padded_masks = []

            for acts, mask in zip(all_activations, all_masks):
                seq_len = acts.shape[1]
                if seq_len < max_seq_len:
                    pad_size = max_seq_len - seq_len
                    acts = torch.cat(
                        [acts, torch.zeros(1, pad_size, hidden_dim, dtype=acts.dtype)],
                        dim=1,
                    )
                    mask = torch.cat(
                        [mask, torch.zeros(1, pad_size, dtype=mask.dtype)], dim=1
                    )
                padded_acts.append(acts)
                padded_masks.append(mask)

            return torch.cat(padded_acts, dim=0), torch.cat(padded_masks, dim=0)

    def get_perplexity(
        self,
        prompts: list[str],
        remote: bool | None = None,
    ) -> np.ndarray:
        """Get perplexity features for prompts (from cache or via extraction).

        Parameters
        ----------
        prompts : list[str]
            Text prompts.
        remote : bool | None
            Override the instance-level remote setting.

        Returns
        -------
        np.ndarray
            Perplexity features with shape (n_prompts, 3).
        """
        if not self.compute_perplexity:
            raise ValueError(
                "UnifiedCache was created with compute_perplexity=False. "
                "Create a new instance with compute_perplexity=True."
            )

        # Ensure cache is warm
        self.warmup(prompts, remote=remote)

        # Load from cache
        features = []
        for prompt in prompts:
            ppl = load_prompt_perplexity(self.model_name, prompt)
            features.append(ppl.float().numpy())

        return np.array(features)

    def get_logits(
        self,
        prompts: list[str],
        remote: bool | None = None,
    ) -> CachedLogits:
        """Get cached logits for prompts (from cache or via extraction).

        Parameters
        ----------
        prompts : list[str]
            Text prompts.
        remote : bool | None
            Override the instance-level remote setting.

        Returns
        -------
        CachedLogits
            Cached logits with values and optional indices tensors.
        """
        if not self.cache_logits:
            raise ValueError(
                "UnifiedCache was created with cache_logits=False. "
                "Create a new instance with cache_logits=True."
            )

        # Ensure cache is warm
        self.warmup(prompts, remote=remote)

        # Load from cache
        all_values = []
        all_indices = []
        for prompt in prompts:
            values, indices = load_prompt_logits(
                self.model_name, prompt, self.logit_top_k
            )
            all_values.append(values)
            if indices is not None:
                all_indices.append(indices)

        return CachedLogits(
            values=torch.cat(all_values, dim=0),
            indices=torch.cat(all_indices, dim=0) if all_indices else None,
            top_k=self.logit_top_k,
            positions=self.logit_positions,
        )



"""Tests for cache backend implementations."""

import time

import pytest

from lmprobe.cache_backends import CacheBackend, LocalCacheBackend


class TestLocalCacheBackend:
    """Tests for LocalCacheBackend (filesystem)."""

    @pytest.fixture
    def backend(self, tmp_path):
        return LocalCacheBackend(tmp_path)

    def test_implements_abc(self, backend):
        assert isinstance(backend, CacheBackend)

    def test_exists_false_when_missing(self, backend):
        assert not backend.exists("nonexistent/key.bin")

    def test_write_and_read_bytes(self, backend):
        backend.write_bytes("model/file.bin", b"hello bytes")
        assert backend.exists("model/file.bin")
        assert backend.read_bytes("model/file.bin") == b"hello bytes"

    def test_write_and_read_text(self, backend):
        backend.write_text("model/_model_name.txt", "my-model")
        assert backend.read_text("model/_model_name.txt") == "my-model"

    def test_delete(self, backend):
        backend.write_bytes("model/file.bin", b"data")
        assert backend.exists("model/file.bin")
        backend.delete("model/file.bin")
        assert not backend.exists("model/file.bin")

    def test_delete_nonexistent_is_noop(self, backend):
        backend.delete("does/not/exist.bin")  # should not raise

    def test_list_keys(self, backend):
        backend.write_bytes("model1/a.bin", b"data")
        backend.write_bytes("model1/b.bin", b"data")
        backend.write_bytes("model2/c.bin", b"data")

        keys = backend.list_keys("model1")
        assert sorted(keys) == ["model1/a.bin", "model1/b.bin"]

    def test_list_keys_empty_prefix(self, backend):
        backend.write_bytes("model1/a.bin", b"data")
        backend.write_bytes("model2/b.bin", b"data")
        keys = backend.list_keys()
        assert len(keys) == 2

    def test_list_keys_nonexistent_prefix(self, backend):
        assert backend.list_keys("nonexistent") == []

    def test_size(self, backend):
        data = b"x" * 100
        backend.write_bytes("model/file.bin", data)
        assert backend.size("model/file.bin") == 100

    def test_mtime(self, backend):
        before = time.time()
        backend.write_bytes("model/file.bin", b"data")
        after = time.time()
        mt = backend.mtime("model/file.bin")
        assert before - 1 <= mt <= after + 1  # allow small tolerance

    def test_touch(self, backend):
        backend.write_bytes("model/file.bin", b"data")
        original_mtime = backend.mtime("model/file.bin")
        time.sleep(0.05)
        backend.touch("model/file.bin")
        new_mtime = backend.mtime("model/file.bin")
        assert new_mtime >= original_mtime

    def test_collect_entries_safetensors(self, backend):
        backend.write_bytes("abc123/prompt1.safetensors", b"data1")
        backend.write_bytes("abc123/prompt2.safetensors", b"data22")

        entries = backend.collect_entries()
        assert len(entries) == 2
        keys = {e[0] for e in entries}
        assert "abc123/prompt1.safetensors" in keys
        assert "abc123/prompt2.safetensors" in keys
        # Each entry is (key, size, mtime)
        for key, size, mtime in entries:
            assert isinstance(size, int)
            assert isinstance(mtime, float)

    def test_collect_entries_skips_metadata(self, backend):
        """_model_name.txt should not appear in collect_entries for v2 files."""
        backend.write_text("abc123/_model_name.txt", "my-model")
        backend.write_bytes("abc123/prompt1.safetensors", b"data")
        entries = backend.collect_entries()
        # Only the safetensors file, not the text file
        assert len(entries) == 1
        assert entries[0][0] == "abc123/prompt1.safetensors"

    def test_delete_tree(self, backend):
        backend.write_bytes("abc123/prompt1.safetensors", b"data")
        backend.write_bytes("abc123/prompt2.safetensors", b"data")
        backend.write_text("abc123/_model_name.txt", "model")

        count = backend.delete_tree("abc123")
        assert count == 2  # two safetensors files (not the _model_name.txt)
        assert not backend.exists("abc123/prompt1.safetensors")
        assert not backend.exists("abc123/_model_name.txt")

    def test_delete_tree_nonexistent(self, backend):
        assert backend.delete_tree("nonexistent") == 0

    def test_write_bytes_creates_parent_dirs(self, backend):
        backend.write_bytes("deep/nested/dir/file.bin", b"data")
        assert backend.read_bytes("deep/nested/dir/file.bin") == b"data"

    def test_write_bytes_atomic(self, backend):
        """No .tmp files left after successful write."""
        backend.write_bytes("model/file.bin", b"data")
        import os

        parent = backend._path("model")
        files = os.listdir(parent)
        assert "file.tmp" not in files
        assert "file.bin" in files



"""Tests for pluggable extraction backend acceptance criteria.

Maps to GitHub issues #19-22:
- #19: ExtractionBackend ABC interface defined, all consumption sites use it
- #20: NnsightBackend wraps existing extraction, caching is backend-agnostic
- #21: LocalBackend uses AutoModelForCausalLM + register_forward_hook
- #22: Device/dtype configuration for LocalBackend
"""

from abc import ABC

import pytest
import torch

from lmprobe import LinearProbe
from lmprobe.backends import (
    ExtractionBackend,
    LocalBackend,
    NnsightBackend,
    _get_decoder_layers,
    resolve_backend,
)

# ── Issue #19: ActivationSource interface ────────────────────────────────────

class TestExtractionBackendABC:
    """Verify ExtractionBackend is a proper ABC with required methods."""

    def test_is_abstract_base_class(self):
        assert issubclass(ExtractionBackend, ABC)

    def test_cannot_instantiate_directly(self):
        """ABC should not be directly instantiatable."""
        with pytest.raises(TypeError, match="abstract"):
            ExtractionBackend("some-model", "cpu")

    def test_has_extract_batch(self):
        assert hasattr(ExtractionBackend, "extract_batch")
        assert getattr(ExtractionBackend.extract_batch, "__isabstractmethod__", False)

    def test_has_extract_batch_with_logits(self):
        assert hasattr(ExtractionBackend, "extract_batch_with_logits")
        assert getattr(
            ExtractionBackend.extract_batch_with_logits,
            "__isabstractmethod__",
            False,
        )

    def test_has_tokenizer_property(self):
        assert isinstance(
            ExtractionBackend.tokenizer, property
        )

    def test_has_model_property(self):
        assert isinstance(
            ExtractionBackend.model, property
        )

    def test_nnsight_implements_interface(self):
        assert issubclass(NnsightBackend, ExtractionBackend)

    def test_local_implements_interface(self):
        assert issubclass(LocalBackend, ExtractionBackend)


class TestAllConsumptionSitesUseBackend:
    """Verify that probe pipeline goes through the backend interface."""

    def test_probe_uses_backend(self, tiny_model):
        """LinearProbe creates an extractor that uses the backend."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            device="cpu",
            backend="local",
        )
        # The extractor should have a backend
        assert hasattr(probe._extractor, "_backend")
        assert isinstance(probe._extractor._backend, ExtractionBackend)
        assert isinstance(probe._extractor._backend, LocalBackend)

    def test_probe_nnsight_backend(self, tiny_model):
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            device="cpu",
            backend="nnsight",
        )
        assert isinstance(probe._extractor._backend, NnsightBackend)


# ── Issue #20: NnsightBackend wraps existing extraction ──────────────────────

class TestNnsightBackend:
    """Verify NnsightBackend properly wraps the existing nnsight code."""

    def test_resolve_nnsight(self, tiny_model):
        backend = resolve_backend("nnsight", tiny_model, "cpu")
        assert isinstance(backend, NnsightBackend)

    def test_nnsight_supports_remote_flag(self, tiny_model):
        backend = resolve_backend("nnsight", tiny_model, "cpu", remote=True)
        assert isinstance(backend, NnsightBackend)
        assert backend.remote is True

    def test_nnsight_default_not_remote(self, tiny_model):
        backend = resolve_backend("nnsight", tiny_model, "cpu")
        assert backend.remote is False

    def test_nnsight_extract_batch(self, tiny_model):
        """NnsightBackend.extract_batch produces correct shapes."""
        backend = resolve_backend("nnsight", tiny_model, "cpu")
        acts, mask = backend.extract_batch(["Hello world"], [0])
        assert acts.ndim == 3  # (batch, seq_len, hidden_dim)
        assert mask.ndim == 2  # (batch, seq_len)
        assert acts.shape[0] == 1

    def test_nnsight_tokenizer_accessible(self, tiny_model):
        backend = resolve_backend("nnsight", tiny_model, "cpu")
        tok = backend.tokenizer
        assert tok is not None
        result = tok("test", return_tensors="pt")
        assert "input_ids" in result


class TestCachingIsBackendAgnostic:
    """Verify that caching works identically for both backends."""

    def test_local_caching(self, tiny_model, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))
        probe = LinearProbe(
            model=tiny_model, layers=-1, device="cpu",
            backend="local", random_state=42,
        )
        probe.fit(["positive"], ["negative"])
        # Cache should have files (safetensors v2 or legacy .pt)
        cache_files = list(tmp_path.rglob("*.safetensors")) + list(tmp_path.rglob("*.pt"))
        assert len(cache_files) > 0

    def test_nnsight_caching(self, tiny_model, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))
        probe = LinearProbe(
            model=tiny_model, layers=-1, device="cpu",
            backend="nnsight", random_state=42,
        )
        probe.fit(["positive"], ["negative"])
        cache_files = list(tmp_path.rglob("*.safetensors")) + list(tmp_path.rglob("*.pt"))
        assert len(cache_files) > 0


# ── Issue #21: LocalBackend with HuggingFace + hooks ────────────────────────

class TestLocalBackendHooks:
    """Verify LocalBackend uses AutoModelForCausalLM and register_forward_hook."""

    def test_model_is_hf_pretrained(self, tiny_model):
        """Model should be a HuggingFace PreTrainedModel."""
        from transformers import PreTrainedModel
        backend = LocalBackend(tiny_model, device="cpu")
        assert isinstance(backend.model, PreTrainedModel)

    def test_decoder_layers_found(self, tiny_model):
        """_get_decoder_layers finds layers for tiny-random-llama."""
        backend = LocalBackend(tiny_model, device="cpu")
        layers = _get_decoder_layers(backend.model)
        assert len(layers) > 0

    def test_hooks_cleaned_up(self, tiny_model):
        """After extraction, no hooks remain on the model."""
        backend = LocalBackend(tiny_model, device="cpu")
        # Force model load
        model = backend.model
        decoder_layers = _get_decoder_layers(model)

        # Count hooks before extraction
        hooks_before = sum(
            len(layer._forward_hooks) for layer in decoder_layers
        )

        # Run extraction
        backend.extract_batch(["test prompt"], [0])

        # Hooks should be cleaned up
        hooks_after = sum(
            len(layer._forward_hooks) for layer in decoder_layers
        )
        assert hooks_after == hooks_before

    def test_extract_single_layer(self, tiny_model):
        backend = LocalBackend(tiny_model, device="cpu")
        acts, mask = backend.extract_batch(["Hello"], [0])
        assert acts.ndim == 3
        assert acts.shape[0] == 1

    def test_extract_multiple_layers_concatenates(self, tiny_model):
        backend = LocalBackend(tiny_model, device="cpu")
        acts_one, _ = backend.extract_batch(["Hello"], [0])
        acts_two, _ = backend.extract_batch(["Hello"], [0, 1])
        assert acts_two.shape[-1] == 2 * acts_one.shape[-1]

    def test_logits_extraction(self, tiny_model):
        backend = LocalBackend(tiny_model, device="cpu")
        acts, mask, logits = backend.extract_batch_with_logits(["Hello"], [0])
        assert logits.ndim == 3  # (batch, seq_len, vocab_size)

    def test_full_pipeline_local(self, tiny_model):
        """Full fit/predict/score pipeline with local backend."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            pooling="last_token",
            classifier="logistic_regression",
            device="cpu",
            backend="local",
            random_state=42,
        )
        positive = ["walk", "fetch", "bark", "wag", "good boy"]
        negative = ["purr", "scratch", "litter", "meow"]
        probe.fit(positive, negative)

        preds = probe.predict(["arf!", "hiss"])
        assert preds.shape == (2,)

        probs = probe.predict_proba(["arf!", "hiss"])
        assert probs.shape == (2, 2)

        acc = probe.score(["arf!", "hiss"], [1, 0])
        assert 0.0 <= acc <= 1.0

    def test_unrecognized_architecture_raises(self):
        """_get_decoder_layers raises for unknown architectures."""
        class FakeModel:
            pass
        with pytest.raises(ValueError, match="Could not find decoder layers"):
            _get_decoder_layers(FakeModel())

    def test_local_remote_incompatible(self, tiny_model):
        with pytest.raises(ValueError, match="does not support remote"):
            resolve_backend("local", tiny_model, "cpu", remote=True)


# ── Issue #22: Device/dtype configuration ────────────────────────────────────

class TestDeviceConfiguration:
    """Verify device parameter works for LocalBackend."""

    def test_cpu_device(self, tiny_model):
        backend = LocalBackend(tiny_model, device="cpu")
        device = next(backend.model.parameters()).device
        assert device.type == "cpu"

    def test_probe_cpu_device(self, tiny_model):
        probe = LinearProbe(
            model=tiny_model, layers=-1, device="cpu", backend="local",
        )
        backend = probe._extractor._backend
        assert isinstance(backend, LocalBackend)
        device = next(backend.model.parameters()).device
        assert device.type == "cpu"

    @pytest.mark.skipif(
        not torch.cuda.is_available()
        or not torch.cuda.get_device_capability()[0] >= 7,
        reason="Compatible CUDA device not available",
    )
    def test_cuda_device(self, tiny_model):
        backend = LocalBackend(tiny_model, device="cuda")
        device = next(backend.model.parameters()).device
        assert device.type == "cuda"
        acts, mask = backend.extract_batch(["Hello"], [0])
        assert acts.device.type == "cuda"

    @pytest.mark.skipif(
        not (hasattr(torch.backends, "mps") and torch.backends.mps.is_available()),
        reason="MPS not available",
    )
    def test_mps_device(self, tiny_model):
        backend = LocalBackend(tiny_model, device="mps")
        device = next(backend.model.parameters()).device
        assert device.type == "mps"
        acts, mask = backend.extract_batch(["Hello"], [0])
        assert acts.device.type == "mps"


class TestDtypeConfiguration:
    """Verify dtype parameter works for LocalBackend."""

    def test_default_dtype_float32(self, tiny_model):
        backend = LocalBackend(tiny_model, device="cpu")
        param_dtype = next(backend.model.parameters()).dtype
        assert param_dtype == torch.float32

    def test_explicit_float32(self, tiny_model):
        backend = LocalBackend(tiny_model, device="cpu", dtype=torch.float32)
        param_dtype = next(backend.model.parameters()).dtype
        assert param_dtype == torch.float32

    def test_bfloat16_dtype(self, tiny_model):
        backend = LocalBackend(tiny_model, device="cpu", dtype=torch.bfloat16)
        param_dtype = next(backend.model.parameters()).dtype
        assert param_dtype == torch.bfloat16

    def test_float16_dtype(self, tiny_model):
        backend = LocalBackend(tiny_model, device="cpu", dtype=torch.float16)
        param_dtype = next(backend.model.parameters()).dtype
        assert param_dtype == torch.float16

    def test_dtype_through_resolve_backend(self, tiny_model):
        backend = resolve_backend(
            "local", tiny_model, "cpu", dtype=torch.bfloat16
        )
        assert isinstance(backend, LocalBackend)
        assert backend.dtype == torch.bfloat16

    def test_dtype_ignored_for_nnsight(self, tiny_model):
        """dtype param should not cause errors for nnsight backend."""
        backend = resolve_backend(
            "nnsight", tiny_model, "cpu", dtype=torch.bfloat16
        )
        assert isinstance(backend, NnsightBackend)

    def test_dtype_through_probe_string(self, tiny_model):
        """LinearProbe accepts dtype as string."""
        probe = LinearProbe(
            model=tiny_model, layers=-1, device="cpu",
            backend="local", dtype="bfloat16",
        )
        backend = probe._extractor._backend
        assert isinstance(backend, LocalBackend)
        assert backend.dtype == torch.bfloat16

    def test_invalid_dtype_string(self, tiny_model):
        with pytest.raises(ValueError, match="Unknown dtype"):
            LinearProbe(
                model=tiny_model, layers=-1, device="cpu",
                backend="local", dtype="invalid",
            )

    def test_bfloat16_extraction_works(self, tiny_model):
        """Activations can be extracted with bfloat16 model."""
        backend = LocalBackend(tiny_model, device="cpu", dtype=torch.bfloat16)
        acts, mask = backend.extract_batch(["Hello"], [0])
        assert acts.ndim == 3
        # Activations should be float (detached from bfloat16 model)
        assert acts.dtype == torch.bfloat16

    def test_bfloat16_full_pipeline(self, tiny_model):
        """Full pipeline works with bfloat16 dtype."""
        probe = LinearProbe(
            model=tiny_model, layers=-1, pooling="last_token",
            device="cpu", backend="local", dtype="bfloat16",
            random_state=42,
        )
        probe.fit(
            ["walk", "fetch", "bark", "wag", "good boy"],
            ["purr", "scratch", "litter", "meow"],
        )
        preds = probe.predict(["arf!", "hiss"])
        assert preds.shape == (2,)

    def test_save_load_preserves_dtype(self, tiny_model, tmp_path):
        """Save/load roundtrip preserves dtype setting."""
        probe = LinearProbe(
            model=tiny_model, layers=-1, device="cpu",
            backend="local", dtype="bfloat16", random_state=42,
        )
        probe.fit(["positive"], ["negative"])

        path = str(tmp_path / "probe.pkl")
        probe.save(path)

        loaded = LinearProbe.load(path)
        assert loaded.dtype == "bfloat16"
        assert loaded.backend == "local"



"""Tests for preprocessing pipeline parameter (Issue #74)."""

from lmprobe import LinearProbe


class TestPreprocessing:
    """Tests for the preprocessing pipeline feature."""

    def test_standard_scaler_string(self, tiny_model):
        """preprocessing='standard' applies StandardScaler."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            preprocessing="standard",
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(["good", "great", "nice"], ["bad", "terrible", "awful"])
        assert probe.preprocessing_pipeline_ is not None
        assert len(probe.preprocessing_pipeline_.steps) == 1
        assert probe.preprocessing_pipeline_.steps[0][0] == "scaler"

    def test_standard_plus_pca(self, tiny_model):
        """preprocessing='standard+pca:4' applies StandardScaler then PCA."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            preprocessing="standard+pca:4",
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(
            ["good", "great", "nice", "wonderful", "excellent"],
            ["bad", "terrible", "awful", "horrible", "dreadful"],
        )
        assert probe.preprocessing_pipeline_ is not None
        assert len(probe.preprocessing_pipeline_.steps) == 2
        assert probe.preprocessing_pipeline_.steps[0][0] == "scaler"
        assert probe.preprocessing_pipeline_.steps[1][0] == "pca"

    def test_pca_with_pca_components_param(self, tiny_model):
        """pca_components param sets PCA n_components when not specified inline."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            preprocessing="standard+pca",
            pca_components=4,
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(
            ["good", "great", "nice", "wonderful", "excellent"],
            ["bad", "terrible", "awful", "horrible", "dreadful"],
        )
        assert probe.preprocessing_pipeline_ is not None
        pca_step = probe.preprocessing_pipeline_.named_steps["pca"]
        assert pca_step.n_components == 4

    def test_fit_predict_with_preprocessing(self, tiny_model):
        """Full roundtrip with preprocessing works correctly."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            preprocessing="standard+pca:4",
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(
            ["good", "great", "nice", "wonderful", "excellent"],
            ["bad", "terrible", "awful", "horrible", "dreadful"],
        )
        predictions = probe.predict(["test input"])
        assert predictions.shape == (1,)

        probabilities = probe.predict_proba(["test input"])
        assert probabilities.shape == (1, 2)

    def test_no_preprocessing_by_default(self, tiny_model):
        """No preprocessing pipeline when preprocessing=None."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(["good", "great"], ["bad", "terrible"])
        assert probe.preprocessing_pipeline_ is None

    def test_preprocessing_as_list(self, tiny_model):
        """preprocessing accepts list of step strings."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            preprocessing=["standard_scaler", "pca:4"],
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(
            ["good", "great", "nice", "wonderful", "excellent"],
            ["bad", "terrible", "awful", "horrible", "dreadful"],
        )
        assert probe.preprocessing_pipeline_ is not None
        assert len(probe.preprocessing_pipeline_.steps) == 2

    def test_invalid_preprocessing_step(self, tiny_model):
        """Unknown preprocessing step raises ValueError."""
        import pytest

        with pytest.raises(ValueError, match="Unknown preprocessing step"):
            probe = LinearProbe(
                model=tiny_model,
                layers=-1,
                preprocessing="unknown_step",
                device="cpu",
                remote=False,
            )
            probe.fit(["good"], ["bad"])

    def test_pca_without_components_raises(self, tiny_model):
        """PCA without component count raises ValueError."""
        import pytest

        with pytest.raises(ValueError, match="PCA requested but no component count"):
            probe = LinearProbe(
                model=tiny_model,
                layers=-1,
                preprocessing="pca",
                device="cpu",
                remote=False,
            )
            probe.fit(["good"], ["bad"])

    def test_normalize_layers_auto_disabled_with_standard(self, tiny_model):
        """normalize_layers auto-disables when preprocessing includes StandardScaler."""
        import warnings

        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            preprocessing="standard",
            normalize_layers=True,
            device="cpu",
            remote=False,
            random_state=42,
        )

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            probe.fit(["good", "great"], ["bad", "terrible"])

        redundancy_warnings = [
            x for x in w if "normalize_layers" in str(x.message)
        ]
        assert len(redundancy_warnings) == 1
        assert "auto-disabled" in str(redundancy_warnings[0].message)

    def test_normalize_layers_false_no_warning_with_standard(self, tiny_model):
        """No warning when normalize_layers=False and preprocessing includes standard."""
        import warnings

        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            preprocessing="standard",
            normalize_layers=False,
            device="cpu",
            remote=False,
            random_state=42,
        )

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            probe.fit(["good", "great"], ["bad", "terrible"])

        redundancy_warnings = [
            x for x in w if "normalize_layers" in str(x.message)
        ]
        assert len(redundancy_warnings) == 0

    def test_normalize_layers_not_disabled_without_standard(self, tiny_model):
        """normalize_layers stays enabled when preprocessing doesn't include standard."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            preprocessing="pca:4",
            normalize_layers=True,
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(
            ["good", "great", "nice", "wonderful", "excellent"],
            ["bad", "terrible", "awful", "horrible", "dreadful"],
        )
        # No auto-disable — scaler_ should exist for single-layer case
        assert hasattr(probe, "scaler_")

    def test_score_with_preprocessing(self, tiny_model):
        """score() works correctly with preprocessing."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            preprocessing="standard+pca:4",
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(
            ["good", "great", "nice", "wonderful", "excellent"],
            ["bad", "terrible", "awful", "horrible", "dreadful"],
        )
        accuracy = probe.score(["test one", "test two"], [1, 0])
        assert 0.0 <= accuracy <= 1.0



"""Tests for mass_mean_augment feature on LinearProbe."""

import numpy as np

from lmprobe import LinearProbe


class TestMassMeanAugmentDefault:
    """Tests that mass_mean_augment=False (default) doesn't change behavior."""

    def test_default_is_false(self, tiny_model):
        """mass_mean_augment defaults to False."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            device="cpu",
            remote=False,
            random_state=42,
        )
        assert probe.mass_mean_augment is False

    def test_no_direction_stored_when_disabled(self, tiny_model):
        """No mass-mean direction is stored when mass_mean_augment=False."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(["good", "great"], ["bad", "terrible"])
        assert probe._mass_mean_direction_ is None


class TestMassMeanAugmentEnabled:
    """Tests for mass_mean_augment=True."""

    def test_direction_stored_after_fit(self, tiny_model):
        """A unit-norm mass-mean direction is stored after fit."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            mass_mean_augment=True,
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(["good", "great"], ["bad", "terrible"])
        assert probe._mass_mean_direction_ is not None
        # Check it's a unit vector
        norm = np.linalg.norm(probe._mass_mean_direction_)
        assert abs(norm - 1.0) < 1e-6

    def test_fit_predict_works(self, tiny_model):
        """fit/predict roundtrip works with mass_mean_augment=True."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            mass_mean_augment=True,
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(
            ["good", "great", "nice"],
            ["bad", "terrible", "awful"],
        )
        predictions = probe.predict(["test input"])
        assert predictions.shape == (1,)
        assert predictions[0] in [0, 1]

    def test_predict_proba_works(self, tiny_model):
        """predict_proba works with mass_mean_augment=True."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            mass_mean_augment=True,
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(
            ["good", "great", "nice"],
            ["bad", "terrible", "awful"],
        )
        probs = probe.predict_proba(["test input"])
        assert probs.shape == (1, 2)
        assert np.allclose(probs.sum(axis=1), 1.0)

    def test_score_works(self, tiny_model):
        """score() works with mass_mean_augment=True."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            mass_mean_augment=True,
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(
            ["good", "great", "nice"],
            ["bad", "terrible", "awful"],
        )
        accuracy = probe.score(["test one", "test two"], [1, 0])
        assert 0.0 <= accuracy <= 1.0


class TestMassMeanAugmentWithPreprocessing:
    """Tests for mass_mean_augment combined with preprocessing."""

    def test_with_standard_plus_pca(self, tiny_model):
        """mass_mean_augment works with standard+pca preprocessing."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            preprocessing="standard+pca:4",
            mass_mean_augment=True,
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(
            ["good", "great", "nice", "wonderful", "excellent"],
            ["bad", "terrible", "awful", "horrible", "dreadful"],
        )
        predictions = probe.predict(["test input"])
        assert predictions.shape == (1,)

        probs = probe.predict_proba(["test input"])
        assert probs.shape == (1, 2)

    def test_without_preprocessing(self, tiny_model):
        """mass_mean_augment works without any preprocessing pipeline."""
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            mass_mean_augment=True,
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe.fit(["good", "great"], ["bad", "terrible"])
        predictions = probe.predict(["test"])
        assert predictions.shape == (1,)


class TestMassMeanAugmentFromActivations:
    """Tests for mass_mean_augment via *_from_activations methods."""

    def test_fit_predict_from_activations(self):
        """mass_mean_augment works with fit/predict_from_activations."""
        probe = LinearProbe(
            mass_mean_augment=True,
            random_state=42,
        )

        rng = np.random.RandomState(42)
        X = rng.randn(20, 10)
        y = np.array([1] * 10 + [0] * 10)

        probe.fit_from_activations(X, y)
        assert probe._mass_mean_direction_ is not None

        preds = probe.predict_from_activations(X)
        assert preds.shape == (20,)

    def test_predict_proba_from_activations(self):
        """predict_proba_from_activations works with mass_mean_augment."""
        probe = LinearProbe(
            mass_mean_augment=True,
            random_state=42,
        )

        rng = np.random.RandomState(42)
        X = rng.randn(20, 10)
        y = np.array([1] * 10 + [0] * 10)

        probe.fit_from_activations(X, y)
        probs = probe.predict_proba_from_activations(X)
        assert probs.shape == (20, 2)
        assert np.allclose(probs.sum(axis=1), 1.0)

    def test_score_from_activations(self):
        """score_from_activations works with mass_mean_augment."""
        probe = LinearProbe(
            mass_mean_augment=True,
            random_state=42,
        )

        rng = np.random.RandomState(42)
        X = rng.randn(20, 10)
        y = np.array([1] * 10 + [0] * 10)

        probe.fit_from_activations(X, y)
        score = probe.score_from_activations(X, y)
        assert 0.0 <= score <= 1.0

    def test_augment_adds_one_feature(self):
        """mass_mean_augment adds exactly one feature column."""
        probe = LinearProbe(
            mass_mean_augment=True,
            random_state=42,
        )

        rng = np.random.RandomState(42)
        n_features = 10
        X = rng.randn(20, n_features)
        y = np.array([1] * 10 + [0] * 10)

        probe.fit_from_activations(X, y)

        # The classifier should have been trained on n_features + 1 columns
        # We verify by checking the coef_ shape if available
        if hasattr(probe.classifier_, "coef_"):
            assert probe.classifier_.coef_.shape[1] == n_features + 1

    def test_disabled_does_not_add_feature(self):
        """mass_mean_augment=False does not add extra feature."""
        probe = LinearProbe(
            mass_mean_augment=False,
            random_state=42,
        )

        rng = np.random.RandomState(42)
        n_features = 10
        X = rng.randn(20, n_features)
        y = np.array([1] * 10 + [0] * 10)

        probe.fit_from_activations(X, y)

        if hasattr(probe.classifier_, "coef_"):
            assert probe.classifier_.coef_.shape[1] == n_features



"""Tests for activation caching (v1 legacy + v2 safetensors)."""

import os

import pytest
import torch

from lmprobe.cache import (
    CacheInfo,
    _hash_string,
    cache_info,
    get_cached_layers,
    get_extraction_cache_dir,
    get_prompt_cache_dir,
    get_prompt_cache_path,
    get_prompt_cached_layers,
    invalidate_extraction_cache,
    is_prompt_fully_cached,
    is_prompt_pooled_cached,
    load_attention_mask,
    load_layer,
    load_prompt_activations,
    load_prompt_pooled_activations,
    save_attention_mask,
    save_layer,
    save_prompt_activations,
    save_prompt_pooled_activations,
    set_cache_dtype,
    set_cache_limit,
)


class TestLegacyCacheStorage:
    """Tests for legacy v1 .pt cache storage functions."""

    def test_save_and_load_layer(self, tmp_path):
        activations = torch.randn(2, 10, 64)
        save_layer(tmp_path, 8, activations)
        loaded = load_layer(tmp_path, 8)
        assert torch.allclose(activations, loaded)

    def test_save_and_load_attention_mask(self, tmp_path):
        mask = torch.ones(2, 10, dtype=torch.long)
        save_attention_mask(tmp_path, mask)
        loaded = load_attention_mask(tmp_path)
        assert torch.equal(mask, loaded)

    def test_get_cached_layers_empty(self, tmp_path):
        assert get_cached_layers(tmp_path) == set()

    def test_get_cached_layers_nonexistent(self, tmp_path):
        assert get_cached_layers(tmp_path / "nonexistent") == set()

    def test_get_cached_layers_finds_layers(self, tmp_path):
        save_layer(tmp_path, 8, torch.randn(2, 10, 64))
        save_layer(tmp_path, 16, torch.randn(2, 10, 64))
        save_layer(tmp_path, 24, torch.randn(2, 10, 64))
        assert get_cached_layers(tmp_path) == {8, 16, 24}

    def test_cache_dir_structure(self):
        cache_dir = get_extraction_cache_dir("test-model", ["prompt1", "prompt2"])
        assert len(cache_dir.parts) >= 4
        assert all(c in "0123456789abcdef" for c in cache_dir.parts[-1])
        assert all(c in "0123456789abcdef" for c in cache_dir.parts[-2])

    def test_invalidate_cache(self, tmp_path):
        save_layer(tmp_path, 8, torch.randn(2, 10, 64))
        save_attention_mask(tmp_path, torch.ones(2, 10))
        invalidate_extraction_cache(tmp_path)
        assert not tmp_path.exists()


class TestSafetensorsFormat:
    """Tests for v2 safetensors cache format (#47)."""

    def test_save_and_load_activations(self, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))
        model = "test-model"
        prompt = "hello world"
        layers = [0, 1]
        acts = torch.randn(1, 5, 128)  # (1, seq_len, 2*hidden_dim)
        mask = torch.ones(1, 5, dtype=torch.long)

        save_prompt_activations(model, prompt, layers, acts, mask)

        # Check safetensors file exists
        sf_path = get_prompt_cache_path(model, prompt)
        assert sf_path.exists()
        assert sf_path.suffix == ".safetensors"

        # Old directory should not exist
        old_dir = get_prompt_cache_dir(model, prompt)
        assert not old_dir.is_dir()

        # Load and verify
        loaded_acts, loaded_mask = load_prompt_activations(model, prompt, layers)
        assert torch.allclose(acts, loaded_acts)
        assert torch.equal(mask, loaded_mask)

    def test_incremental_layer_save(self, tmp_path, monkeypatch):
        """Adding layers to an existing cache file works."""
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))
        model = "test-model"
        prompt = "test prompt"

        # Save layer 0
        acts0 = torch.randn(1, 5, 64)
        mask = torch.ones(1, 5, dtype=torch.long)
        save_prompt_activations(model, prompt, [0], acts0, mask)

        # Save layer 1 (should merge into same file)
        acts1 = torch.randn(1, 5, 64)
        save_prompt_activations(model, prompt, [1], acts1, mask)

        # Both layers should be cached
        assert is_prompt_fully_cached(model, prompt, {0, 1})

        # Load individual layers
        loaded_acts0, _ = load_prompt_activations(model, prompt, [0])
        assert torch.allclose(acts0, loaded_acts0)

        loaded_acts1, _ = load_prompt_activations(model, prompt, [1])
        assert torch.allclose(acts1, loaded_acts1)

    def test_get_prompt_cached_layers_safetensors(self, tmp_path, monkeypatch):
        """get_prompt_cached_layers works with safetensors format."""
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))
        model = "test-model"
        prompt = "test prompt"

        acts = torch.randn(1, 5, 128)
        mask = torch.ones(1, 5, dtype=torch.long)
        save_prompt_activations(model, prompt, [0, 1], acts, mask)

        cache_dir = get_prompt_cache_dir(model, prompt)
        cached = get_prompt_cached_layers(cache_dir)
        assert cached == {0, 1}

    def test_invalidation_deletes_safetensors(self, tmp_path, monkeypatch):
        """invalidate_extraction_cache removes safetensors files too."""
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))
        model = "test-model"
        prompt = "test"

        save_prompt_activations(
            model, prompt, [0],
            torch.randn(1, 5, 64),
            torch.ones(1, 5, dtype=torch.long),
        )

        cache_dir = get_prompt_cache_dir(model, prompt)
        invalidate_extraction_cache(cache_dir)

        assert not get_prompt_cache_path(model, prompt).exists()
        assert not is_prompt_fully_cached(model, prompt, {0})

    def test_model_name_registered(self, tmp_path, monkeypatch):
        """Model name is recorded for cache_info()."""
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))
        model = "org/my-model"

        save_prompt_activations(
            model, "test", [0],
            torch.randn(1, 5, 64),
            torch.ones(1, 5, dtype=torch.long),
        )

        model_hash = _hash_string(model)
        name_file = tmp_path / model_hash / "_model_name.txt"
        assert name_file.exists()
        assert name_file.read_text().strip() == model


class TestPooledCacheSafetensors:
    """Tests for pooled cache in safetensors format."""

    def test_save_and_load_pooled(self, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))
        model = "test-model"
        prompt = "test"
        layers = [0, 1]
        pooling = "last_token"
        pooled = torch.randn(1, 128)

        save_prompt_pooled_activations(model, prompt, layers, pooled, pooling)
        assert is_prompt_pooled_cached(model, prompt, {0, 1}, pooling)

        loaded = load_prompt_pooled_activations(model, prompt, layers, pooling)
        assert torch.allclose(pooled, loaded)

    def test_raw_and_pooled_coexist(self, tmp_path, monkeypatch):
        """Raw and pooled activations coexist in the same file."""
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))
        model = "test-model"
        prompt = "test"

        # Save raw
        save_prompt_activations(
            model, prompt, [0],
            torch.randn(1, 5, 64),
            torch.ones(1, 5, dtype=torch.long),
        )
        # Save pooled
        save_prompt_pooled_activations(
            model, prompt, [0], torch.randn(1, 64), "last_token"
        )

        # Both should be cached
        assert is_prompt_fully_cached(model, prompt, {0})
        assert is_prompt_pooled_cached(model, prompt, {0}, "last_token")

        # Only one file on disk
        sf_path = get_prompt_cache_path(model, prompt)
        assert sf_path.exists()


class TestDiskFullError:
    """Tests for disk-full error handling (#44)."""

    def test_clear_error_message_on_disk_full(self, tmp_path, monkeypatch):
        """Disk-full errors are wrapped with a helpful message."""
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        def fake_save_file(*args, **kwargs):
            raise OSError("No space left on device")

        monkeypatch.setattr("safetensors.torch.save_file", fake_save_file)

        with pytest.raises(OSError, match="Disk full"):
            save_prompt_activations(
                "test", "test", [0],
                torch.randn(1, 3, 64),
                torch.ones(1, 3, dtype=torch.long),
            )


class TestCacheDtype:
    """Tests for float16 cache storage (#46)."""

    def test_float16_storage(self, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))
        set_cache_dtype("float16")
        try:
            acts = torch.randn(1, 5, 64, dtype=torch.float32)
            mask = torch.ones(1, 5, dtype=torch.long)
            save_prompt_activations("test", "fp16test", [0], acts, mask)

            loaded_acts, loaded_mask = load_prompt_activations("test", "fp16test", [0])
            # Activations should be stored as float16
            assert loaded_acts.dtype == torch.float16
            # Mask should remain long
            assert loaded_mask.dtype == torch.int64
            # Values should be close (within float16 precision)
            assert torch.allclose(acts.half(), loaded_acts)
        finally:
            set_cache_dtype(None)

    def test_env_var_dtype(self, tmp_path, monkeypatch):
        """set_cache_dtype changes the active dtype and saves respect it."""
        import lmprobe.cache as cache_mod

        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        set_cache_dtype("float16")
        try:
            assert cache_mod._CACHE_DTYPE == torch.float16

            acts = torch.randn(1, 3, 64, dtype=torch.float32)
            mask = torch.ones(1, 3, dtype=torch.long)
            save_prompt_activations("test", "dtype-check", [0], acts, mask)

            loaded_acts, _ = load_prompt_activations("test", "dtype-check", [0])
            assert loaded_acts.dtype == torch.float16
        finally:
            set_cache_dtype(None)


class TestCacheInfo:
    """Tests for cache_info() reporting (#48)."""

    def test_empty_cache(self, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))
        info = cache_info()
        assert isinstance(info, CacheInfo)
        assert info.total_size_bytes == 0
        assert len(info.models) == 0

    def test_cache_info_reports_model(self, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        save_prompt_activations(
            "test-model", "prompt1", [0, 1],
            torch.randn(1, 5, 128),
            torch.ones(1, 5, dtype=torch.long),
        )
        save_prompt_activations(
            "test-model", "prompt2", [0, 1],
            torch.randn(1, 5, 128),
            torch.ones(1, 5, dtype=torch.long),
        )

        info = cache_info()
        assert info.total_size_bytes > 0
        assert len(info.models) == 1
        assert info.models[0].model_name == "test-model"
        assert info.models[0].num_prompts == 2
        assert info.models[0].num_layers == 2

    def test_cache_info_filter_by_model(self, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        save_prompt_activations(
            "model-a", "p1", [0], torch.randn(1, 3, 64),
            torch.ones(1, 3, dtype=torch.long),
        )
        save_prompt_activations(
            "model-b", "p1", [0], torch.randn(1, 3, 64),
            torch.ones(1, 3, dtype=torch.long),
        )

        info_a = cache_info(model="model-a")
        assert len(info_a.models) == 1
        assert info_a.models[0].model_name == "model-a"

    def test_cache_info_repr(self, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))
        info = cache_info()
        text = repr(info)
        assert "Cache directory" in text
        assert "Total size" in text


class TestLRUEviction:
    """Tests for LRU eviction (#49)."""

    def test_eviction_respects_limit(self, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        # Save several prompts
        for i in range(5):
            save_prompt_activations(
                "test-model", f"prompt-{i}", [0],
                torch.randn(1, 10, 64),
                torch.ones(1, 10, dtype=torch.long),
            )

        # Get total size
        info = cache_info()
        total_bytes = info.total_size_bytes
        assert total_bytes > 0

        # Set limit to half the current size
        half = total_bytes / 2
        set_cache_limit(gb=half / (1024**3))

        try:
            # Trigger eviction by saving one more
            save_prompt_activations(
                "test-model", "prompt-trigger", [0],
                torch.randn(1, 10, 64),
                torch.ones(1, 10, dtype=torch.long),
            )

            # Cache should now be near the limit (allow for the trigger prompt)
            info_after = cache_info()
            one_prompt_size = total_bytes / 5  # approximate size of one prompt
            assert info_after.total_size_bytes <= half + one_prompt_size
        finally:
            set_cache_limit(None)

    def test_set_cache_limit_zero_disables(self, tmp_path, monkeypatch):
        """LMPROBE_CACHE_MAX_GB=0 disables caching."""
        import lmprobe.cache as cache_mod

        old_val = cache_mod._CACHE_MAX_BYTES
        try:
            set_cache_limit(gb=0)
            assert cache_mod._CACHE_MAX_BYTES == -1
        finally:
            cache_mod._CACHE_MAX_BYTES = old_val


class TestCachedExtractor:
    """Tests for CachedExtractor with safetensors caching."""

    def test_caches_prompts_individually(self, tiny_model, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe.cache import CachedExtractor
        from lmprobe.extraction import ActivationExtractor

        extractor = ActivationExtractor(
            tiny_model, device="cpu", layers=[0, 1], batch_size=4,
        )
        cached = CachedExtractor(extractor)
        cached.extract(["hello world"], remote=False)

        # Check cache
        cache_dir = get_prompt_cache_dir(tiny_model, "hello world")
        cached_layers = get_prompt_cached_layers(cache_dir)
        assert cached_layers == {0, 1}

    def test_full_cache_hit(self, tiny_model, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe.cache import CachedExtractor
        from lmprobe.extraction import ActivationExtractor

        extractor = ActivationExtractor(
            tiny_model, device="cpu", layers=[0, 1], batch_size=4,
        )
        cached = CachedExtractor(extractor)

        acts1, mask1 = cached.extract(["cached prompt"], remote=False)
        acts2, mask2 = cached.extract(["cached prompt"], remote=False)

        assert torch.allclose(acts1, acts2)
        assert torch.equal(mask1, mask2)

    def test_invalidate_cache_forces_reextraction(self, tiny_model, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe.cache import CachedExtractor
        from lmprobe.extraction import ActivationExtractor

        extractor = ActivationExtractor(
            tiny_model, device="cpu", layers=[0], batch_size=4,
        )
        cached = CachedExtractor(extractor)
        cached.extract(["invalidation test"], remote=False)

        sf_path = get_prompt_cache_path(tiny_model, "invalidation test")
        assert sf_path.exists()

        cached.extract(["invalidation test"], remote=False, invalidate_cache=True)
        assert sf_path.exists()

    def test_cross_request_cache_hit(self, tiny_model, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe.cache import CachedExtractor
        from lmprobe.extraction import ActivationExtractor

        extractor = ActivationExtractor(
            tiny_model, device="cpu", layers=[0, 1], batch_size=4,
        )
        cached = CachedExtractor(extractor)

        cached.extract(["prompt A", "prompt B"], remote=False)
        cached.extract(["prompt B", "prompt C"], remote=False)

        assert is_prompt_fully_cached(tiny_model, "prompt A", {0, 1})
        assert is_prompt_fully_cached(tiny_model, "prompt B", {0, 1})
        assert is_prompt_fully_cached(tiny_model, "prompt C", {0, 1})


class TestLinearProbeWithCache:
    """Integration tests for LinearProbe with caching."""

    def test_iterative_layer_experimentation(self, tiny_model, tmp_path, monkeypatch):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe import LinearProbe

        positive = ["good example"]
        negative = ["bad example"]

        probe1 = LinearProbe(model=tiny_model, layers=0, device="cpu", remote=False)
        probe1.fit(positive, negative)

        for prompt in positive + negative:
            cache_dir = get_prompt_cache_dir(tiny_model, prompt)
            assert get_prompt_cached_layers(cache_dir) == {0}

        probe2 = LinearProbe(model=tiny_model, layers=[0, 1], device="cpu", remote=False)
        probe2.fit(positive, negative)

        assert probe1.classifier_ is not None
        assert probe2.classifier_ is not None

        for prompt in positive + negative:
            cache_dir = get_prompt_cache_dir(tiny_model, prompt)
            assert get_prompt_cached_layers(cache_dir) == {0, 1}


class TestUnifiedCacheLinearProbeIntegration:
    """Tests for cache compatibility between UnifiedCache and LinearProbe."""

    def test_unified_cache_warmup_enables_linear_probe_cache_hit(
        self, tiny_model, tmp_path, monkeypatch
    ):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe import LinearProbe, UnifiedCache
        from lmprobe.cache import is_prompt_fully_cached

        prompts = ["This is a test prompt", "Another test"]
        layers = [0, 1]

        cache = UnifiedCache(
            model=tiny_model, layers=layers,
            compute_perplexity=False, device="cpu", remote=False,
            cache_pooled=False,
        )
        stats = cache.warmup(prompts)
        assert stats.activations_extracted == len(prompts)

        for prompt in prompts:
            assert is_prompt_fully_cached(tiny_model, prompt, set(layers))

        probe = LinearProbe(model=tiny_model, layers=layers, device="cpu", remote=False)
        probe.fit([prompts[0]], [prompts[1]])

        for prompt in prompts:
            cache_dir = get_prompt_cache_dir(tiny_model, prompt)
            cached = get_prompt_cached_layers(cache_dir)
            assert cached == set(layers)

    def test_pooled_cache_warmup_enables_linear_probe_cache_hit(
        self, tiny_model, tmp_path, monkeypatch
    ):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe import LinearProbe, UnifiedCache
        from lmprobe.cache import is_prompt_pooled_cached

        prompts = ["This is a test prompt", "Another test"]
        layers = [0, 1]
        pooling = "last_token"

        cache = UnifiedCache(
            model=tiny_model, layers=layers,
            compute_perplexity=False, device="cpu", remote=False,
            cache_pooled=True, pooling=pooling,
        )
        stats = cache.warmup(prompts)
        assert stats.activations_extracted == len(prompts)

        for prompt in prompts:
            assert is_prompt_pooled_cached(tiny_model, prompt, set(layers), pooling)

        probe = LinearProbe(
            model=tiny_model, layers=layers, pooling=pooling,
            device="cpu", remote=False,
        )

        cache_files_before = sum(len(files) for _, _, files in os.walk(tmp_path))
        probe.fit([prompts[0]], [prompts[1]])
        cache_files_after = sum(len(files) for _, _, files in os.walk(tmp_path))

        assert cache_files_after == cache_files_before, (
            f"LinearProbe should not create new cache files when pooled cache exists. "
            f"Before: {cache_files_before}, After: {cache_files_after}"
        )

    def test_linear_probe_respects_pooled_cache_format(
        self, tiny_model, tmp_path, monkeypatch
    ):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe import LinearProbe, UnifiedCache

        prompts = ["Test prompt one", "Test prompt two"]
        layers = [0, 1]
        pooling = "last_token"

        cache = UnifiedCache(
            model=tiny_model, layers=layers,
            compute_perplexity=False, device="cpu", remote=False,
            cache_pooled=True, pooling=pooling,
        )
        cache.warmup(prompts)
        acts_from_cache, _ = cache.get_activations(prompts)

        probe = LinearProbe(
            model=tiny_model, layers=layers, pooling=pooling,
            device="cpu", remote=False,
        )
        probe.fit([prompts[0]], [prompts[1]])

        from lmprobe.pooling import get_pooling_fn

        pooling_fn = get_pooling_fn(pooling)
        acts_raw, mask = probe._cached_extractor.extract(prompts, remote=False)
        acts_from_probe = pooling_fn(acts_raw, mask)

        assert torch.allclose(acts_from_cache, acts_from_probe, atol=1e-5)


class TestUnifiedCacheActivationBaselineIntegration:
    """Tests for cache compatibility between UnifiedCache and ActivationBaseline."""

    def test_pooled_cache_warmup_enables_activation_baseline_cache_hit(
        self, tiny_model, tmp_path, monkeypatch
    ):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe import ActivationBaseline, UnifiedCache
        from lmprobe.cache import is_prompt_pooled_cached

        prompts = ["This is a test prompt", "Another test"]
        layers = [0, 1]
        pooling = "last_token"

        cache = UnifiedCache(
            model=tiny_model, layers=layers,
            compute_perplexity=False, device="cpu", remote=False,
            cache_pooled=True, pooling=pooling,
        )
        stats = cache.warmup(prompts)
        assert stats.activations_extracted == len(prompts)

        for prompt in prompts:
            assert is_prompt_pooled_cached(tiny_model, prompt, set(layers), pooling)

        cache_files_before = sum(len(files) for _, _, files in os.walk(tmp_path))

        baseline = ActivationBaseline(
            model=tiny_model, method="random_direction",
            layers=layers, pooling=pooling,
            device="cpu", remote=False, random_state=42,
        )
        baseline.fit([prompts[0]], [prompts[1]])

        cache_files_after = sum(len(files) for _, _, files in os.walk(tmp_path))
        assert cache_files_after == cache_files_before

    def test_pooled_cache_works_for_pca_baseline(
        self, tiny_model, tmp_path, monkeypatch
    ):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe import ActivationBaseline, UnifiedCache

        prompts = ["Test prompt one", "Test prompt two", "Test prompt three"]
        layers = [0, 1]
        pooling = "last_token"

        cache = UnifiedCache(
            model=tiny_model, layers=layers,
            compute_perplexity=False, device="cpu", remote=False,
            cache_pooled=True, pooling=pooling,
        )
        cache.warmup(prompts)

        cache_files_before = sum(len(files) for _, _, files in os.walk(tmp_path))

        baseline = ActivationBaseline(
            model=tiny_model, method="pca", layers=layers,
            pooling=pooling, n_components=2, device="cpu", remote=False,
        )
        baseline.fit([prompts[0]], [prompts[1]])
        baseline.predict([prompts[2]])

        cache_files_after = sum(len(files) for _, _, files in os.walk(tmp_path))
        assert cache_files_after == cache_files_before

    def test_layer_0_baseline_uses_correct_cache(
        self, tiny_model, tmp_path, monkeypatch
    ):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe import ActivationBaseline, UnifiedCache
        from lmprobe.cache import is_prompt_pooled_cached

        prompts = ["Test prompt A", "Test prompt B"]
        pooling = "last_token"

        cache = UnifiedCache(
            model=tiny_model, layers=[0], compute_perplexity=False,
            device="cpu", remote=False, cache_pooled=True, pooling=pooling,
        )
        cache.warmup(prompts)

        for prompt in prompts:
            assert is_prompt_pooled_cached(tiny_model, prompt, {0}, pooling)

        cache_files_before = sum(len(files) for _, _, files in os.walk(tmp_path))

        baseline = ActivationBaseline(
            model=tiny_model, method="layer_0", pooling=pooling,
            device="cpu", remote=False,
        )
        baseline.fit([prompts[0]], [prompts[1]])

        cache_files_after = sum(len(files) for _, _, files in os.walk(tmp_path))
        assert cache_files_after == cache_files_before


class TestUnifiedCachePerplexityBaselineIntegration:
    """Tests for perplexity cache compatibility."""

    def test_perplexity_cache_warmup_enables_baseline_cache_hit(
        self, tiny_model, tmp_path, monkeypatch
    ):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe import BaselineProbe, UnifiedCache
        from lmprobe.cache import is_prompt_perplexity_cached

        prompts = ["This is a test prompt", "Another test"]

        cache = UnifiedCache(
            model=tiny_model, layers=[0, 1],
            compute_perplexity=True, device="cpu", remote=False,
            cache_pooled=True,
        )
        stats = cache.warmup(prompts)
        assert stats.perplexity_extracted == len(prompts)

        for prompt in prompts:
            assert is_prompt_perplexity_cached(tiny_model, prompt)

        cache_files_before = sum(len(files) for _, _, files in os.walk(tmp_path))

        baseline = BaselineProbe(
            method="perplexity", model=tiny_model, device="cpu", remote=False,
        )
        baseline.fit([prompts[0]], [prompts[1]])

        cache_files_after = sum(len(files) for _, _, files in os.walk(tmp_path))
        assert cache_files_after == cache_files_before

    def test_perplexity_cache_shared_between_unified_and_baseline(
        self, tiny_model, tmp_path, monkeypatch
    ):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        from lmprobe import BaselineProbe, UnifiedCache
        from lmprobe.cache import is_prompt_perplexity_cached

        prompts = ["Forward direction test", "Backward direction test"]

        baseline = BaselineProbe(
            method="perplexity", model=tiny_model, device="cpu", remote=False,
        )
        baseline.fit([prompts[0]], [prompts[1]])

        for prompt in prompts:
            assert is_prompt_perplexity_cached(tiny_model, prompt)

        cache = UnifiedCache(
            model=tiny_model, layers=[0, 1],
            compute_perplexity=True, device="cpu", remote=False,
        )
        stats = cache.warmup(prompts)

        assert stats.perplexity_cached == len(prompts)
        assert stats.perplexity_extracted == 0

    def test_unified_cache_perplexity_matches_baseline(
        self, tiny_model, tmp_path, monkeypatch
    ):
        monkeypatch.setenv("LMPROBE_CACHE_DIR", str(tmp_path))

        import numpy as np

        from lmprobe import BaselineProbe, UnifiedCache

        prompts = ["Matching test prompt"]

        cache = UnifiedCache(
            model=tiny_model, layers=[0],
            compute_perplexity=True, device="cpu", remote=False,
        )
        cache.warmup(prompts)
        ppl_from_cache = cache.get_perplexity(prompts)

        baseline = BaselineProbe(
            method="perplexity", model=tiny_model, device="cpu", remote=False,
        )
        ppl_from_baseline = baseline._compute_perplexity(prompts)

        assert np.allclose(ppl_from_cache, ppl_from_baseline, atol=1e-5)



"""
North Star Test: The README example must run exactly as documented.

This test runs the exact code from README.md's "Example Usage" section.
If this test passes, the library's public API is working as advertised.

This is the PRIMARY test for the library. All other tests support this one.
"""

import numpy as np


def test_readme_example_runs(tiny_model):
    """The README example code runs without error.

    This test mirrors the exact usage pattern from README.md,
    substituting the tiny test model for the full Llama model.
    """
    from lmprobe import LinearProbe

    positive_prompts = [
        "Who wants to go for a walk?",
        "My tail is wagging with delight.",
        "Fetch the ball!",
        "Good boy!",
        "Slobbering, chewing, growling, barking.",
    ]

    negative_prompts = [
        "Enjoys lounging in the sun beam all day.",
        "Purring, stalking, pouncing, scratching.",
        "Uses a litterbox, throws sand all over the room.",
        "Tail raised, back arched, eyes alert, whiskers forward.",
    ]

    # Configure the probe (mirrors README, with test-appropriate values)
    probe = LinearProbe(
        model=tiny_model,
        layers=-1,  # Last layer (tiny model has few layers)
        pooling="last_token",
        classifier="logistic_regression",
        device="cpu",
        remote=False,
        random_state=42,
    )

    # Fit using contrastive prompts
    probe.fit(positive_prompts, negative_prompts)

    # Predict on new examples
    test_prompts = [
        "Arf! Arf! Let's go outside!",
        "Knocking things off the counter for sport.",
    ]
    predictions = probe.predict(test_prompts)
    probabilities = probe.predict_proba(test_prompts)

    # Verify output shapes (values are meaningless with random weights)
    assert isinstance(predictions, np.ndarray)
    assert predictions.shape == (2,)
    assert set(predictions).issubset({0, 1})  # Binary predictions

    assert isinstance(probabilities, np.ndarray)
    assert probabilities.shape == (2, 2)  # (n_samples, n_classes)
    assert np.allclose(probabilities.sum(axis=1), 1.0)  # Probabilities sum to 1

    # Evaluate
    accuracy = probe.score(test_prompts, [1, 0])
    assert isinstance(accuracy, float)
    assert 0.0 <= accuracy <= 1.0


def test_readme_save_load(tiny_model, tmp_path):
    """The save/load functionality from README works."""
    from lmprobe import LinearProbe

    probe = LinearProbe(
        model=tiny_model,
        layers=-1,
        pooling="last_token",
        device="cpu",
        remote=False,
        random_state=42,
    )

    probe.fit(["positive"], ["negative"])

    # Save/load for deployment
    save_path = tmp_path / "test_probe.pkl"
    probe.save(str(save_path))

    loaded_probe = LinearProbe.load(str(save_path))

    # Loaded probe should produce same predictions
    original_pred = probe.predict(["test input"])
    loaded_pred = loaded_probe.predict(["test input"])

    assert np.array_equal(original_pred, loaded_pred)


def test_readme_different_pooling(tiny_model):
    """Different train vs inference pooling from README works."""
    from lmprobe import LinearProbe

    probe = LinearProbe(
        model=tiny_model,
        layers=-1,
        pooling="last_token",
        inference_pooling="max",  # Override for predict()
        device="cpu",
        remote=False,
        random_state=42,
    )

    probe.fit(["positive example"], ["negative example"])
    predictions = probe.predict(["test input"])

    assert predictions.shape == (1,)


def test_readme_per_token_scores(tiny_model):
    """Per-token scoring with inference_pooling='all' works."""
    from lmprobe import LinearProbe

    probe = LinearProbe(
        model=tiny_model,
        layers=-1,
        pooling="last_token",
        inference_pooling="all",  # Return per-token scores
        device="cpu",
        remote=False,
        random_state=42,
    )

    probe.fit(["positive example"], ["negative example"])

    # With inference_pooling="all", we get per-token probabilities
    token_probs = probe.predict_proba(["test input with multiple tokens"])

    # Shape should be (batch, seq_len, n_classes) or (batch, seq_len) for binary
    assert len(token_probs.shape) >= 2
    assert token_probs.shape[0] == 1  # batch size


def test_readme_multilayer(tiny_model):
    """Multi-layer probing (concatenation) works."""
    from lmprobe import LinearProbe

    probe = LinearProbe(
        model=tiny_model,
        layers=[-2, -1],  # Last two layers, concatenated
        pooling="last_token",
        device="cpu",
        remote=False,
        random_state=42,
    )

    probe.fit(["positive"], ["negative"])
    predictions = probe.predict(["test"])

    assert predictions.shape == (1,)



"""Tests for HuggingFace Hub integration."""

from __future__ import annotations

import json
from pathlib import Path

import numpy as np
import pytest

from lmprobe import LinearProbe

TEST_MODEL = "stas/tiny-random-llama-2"

try:
    import skops  # noqa: F401

    HAS_SKOPS = True
except ImportError:
    HAS_SKOPS = False

requires_skops = pytest.mark.skipif(not HAS_SKOPS, reason="skops not installed")


@pytest.fixture
def fitted_probe():
    """Return a fitted probe using the tiny model."""
    probe = LinearProbe(
        model=TEST_MODEL,
        layers=-1,
        pooling="last_token",
        classifier="logistic_regression",
        device="cpu",
        remote=False,
        random_state=42,
    )
    positive = ["dog walks bark fetch", "good boy wag tail"]
    negative = ["cat purr scratch meow", "kitty litter nap"]
    probe.fit(positive, negative)
    return probe


@pytest.fixture
def tiny_model():
    return TEST_MODEL


class TestTrainingDataCaching:
    """Test that fit() caches training prompts."""

    def test_contrastive_mode_caches_prompts(self, fitted_probe):
        assert fitted_probe._training_positive_ == [
            "dog walks bark fetch",
            "good boy wag tail",
        ]
        assert fitted_probe._training_negative_ == [
            "cat purr scratch meow",
            "kitty litter nap",
        ]
        assert fitted_probe._training_prompts_ is None
        assert fitted_probe._training_labels_ is None

    def test_standard_mode_caches_prompts(self, tiny_model):
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            pooling="last_token",
            device="cpu",
            remote=False,
            random_state=42,
        )
        prompts = ["hello world", "foo bar", "baz qux"]
        labels = [1, 0, 1]
        probe.fit(prompts, labels)

        assert probe._training_prompts_ == prompts
        assert probe._training_labels_ == labels
        assert probe._training_positive_ is None
        assert probe._training_negative_ is None

    def test_unfitted_probe_has_no_cached_data(self, tiny_model):
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            device="cpu",
            remote=False,
        )
        assert probe._training_positive_ is None
        assert probe._training_negative_ is None
        assert probe._training_prompts_ is None
        assert probe._training_labels_ is None


class TestEvaluate:
    """Test the evaluate() method."""

    def test_evaluate_returns_metrics(self, fitted_probe):
        test_prompts = ["woof bark play", "meow purr sleep"]
        test_labels = [1, 0]
        results = fitted_probe.evaluate(test_prompts, test_labels)

        assert "accuracy" in results
        assert "f1" in results
        assert "precision" in results
        assert "recall" in results
        assert "n_eval" in results
        assert "eval_hash" in results
        assert results["n_eval"] == 2
        assert 0.0 <= results["accuracy"] <= 1.0

    def test_evaluate_includes_auroc(self, fitted_probe):
        test_prompts = ["woof bark play", "meow purr sleep"]
        test_labels = [1, 0]
        results = fitted_probe.evaluate(test_prompts, test_labels)

        # LogisticRegression supports predict_proba, so auroc should be present
        assert "auroc" in results
        assert 0.0 <= results["auroc"] <= 1.0

    def test_evaluate_caches_results(self, fitted_probe):
        test_prompts = ["woof bark play", "meow purr sleep"]
        test_labels = [1, 0]
        results = fitted_probe.evaluate(test_prompts, test_labels)

        assert fitted_probe._evaluation_results_ is results

    def test_evaluate_on_unfitted_raises(self, tiny_model):
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            device="cpu",
            remote=False,
        )
        with pytest.raises(RuntimeError, match="not been fitted"):
            probe.evaluate(["test"], [1])


class TestBuildProbeConfig:
    """Test _build_probe_config."""

    def test_config_has_required_fields(self, fitted_probe):
        from lmprobe.hub import _build_probe_config

        config = _build_probe_config(fitted_probe)

        assert "lmprobe_version" in config
        assert "config_version" in config
        assert config["config_version"] == 1
        assert "base_model" in config
        assert config["base_model"]["name"] == TEST_MODEL
        assert "probe" in config
        assert "classes" in config

    def test_config_probe_section(self, fitted_probe):
        from lmprobe.hub import _build_probe_config

        config = _build_probe_config(fitted_probe)
        probe_cfg = config["probe"]

        assert probe_cfg["layers"] is not None
        assert probe_cfg["pooling"] == "last_token"
        assert probe_cfg["classifier_type"] == "logistic_regression"
        assert probe_cfg["task"] == "classification"
        assert probe_cfg["random_state"] == 42

    def test_config_with_class_labels(self, fitted_probe):
        from lmprobe.hub import _build_probe_config

        config = _build_probe_config(
            fitted_probe,
            class_labels={0: "negative", 1: "positive"},
        )
        assert config["class_labels"] == {"0": "negative", "1": "positive"}

    def test_config_classes_from_fitted(self, fitted_probe):
        from lmprobe.hub import _build_probe_config

        config = _build_probe_config(fitted_probe)
        assert config["classes"] == [0, 1]


class TestBuildTrainingInfo:
    """Test _build_training_info."""

    def test_training_info_with_data(self, fitted_probe):
        from lmprobe.hub import _build_training_info

        info = _build_training_info(fitted_probe, include_training_data=True)

        td = info["training_data"]
        assert td["n_positive"] == 2
        assert td["n_negative"] == 2
        assert "positive_hash" in td
        assert td["positive_hash"].startswith("sha256:")
        assert "positive_examples" in td
        assert "negative_examples" in td

    def test_training_info_without_data(self, fitted_probe):
        from lmprobe.hub import _build_training_info

        info = _build_training_info(fitted_probe, include_training_data=False)

        td = info["training_data"]
        assert td["n_positive"] == 2
        assert "positive_hash" in td
        assert "positive_examples" not in td
        assert "negative_examples" not in td

    def test_training_info_with_metrics(self, fitted_probe):
        from lmprobe.hub import _build_training_info

        metrics = {"accuracy": 0.95, "auroc": 0.98, "n_eval": 20, "eval_hash": "sha256:abc"}
        info = _build_training_info(fitted_probe, metrics=metrics)

        assert info["evaluation"]["metrics"]["accuracy"] == 0.95
        assert info["evaluation"]["metrics"]["auroc"] == 0.98
        assert info["evaluation"]["eval_set_size"] == 20

    def test_training_info_environment(self, fitted_probe):
        from lmprobe.hub import _build_training_info

        info = _build_training_info(fitted_probe)
        env = info["training_environment"]

        assert "lmprobe_version" in env
        assert "python_version" in env
        assert "torch_version" in env
        assert "sklearn_version" in env

    def test_training_info_timestamps(self, fitted_probe):
        from lmprobe.hub import _build_training_info

        info = _build_training_info(fitted_probe)
        assert "timestamps" in info
        assert "trained_at" in info["timestamps"]
        assert "pushed_at" in info["timestamps"]


class TestRenderModelCard:
    """Test _render_model_card."""

    def test_model_card_has_yaml_frontmatter(self, fitted_probe):
        from lmprobe.hub import _build_probe_config, _build_training_info, _render_model_card

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        info = _build_training_info(fitted_probe)
        card = _render_model_card(config, info)

        assert card.startswith("---\n")
        assert "library_name: lmprobe" in card
        assert "pipeline_tag: text-classification" in card

    def test_model_card_with_description(self, fitted_probe):
        from lmprobe.hub import _build_probe_config, _build_training_info, _render_model_card

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        info = _build_training_info(fitted_probe)
        card = _render_model_card(config, info, description="Detects dogs vs cats")

        assert "Detects dogs vs cats" in card

    def test_model_card_with_tags(self, fitted_probe):
        from lmprobe.hub import _build_probe_config, _build_training_info, _render_model_card

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        info = _build_training_info(fitted_probe)
        card = _render_model_card(config, info, tags=["safety", "custom-tag"])

        assert "  - safety" in card
        assert "  - custom-tag" in card

    def test_model_card_no_evaluation(self, fitted_probe):
        from lmprobe.hub import _build_probe_config, _build_training_info, _render_model_card

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        info = _build_training_info(fitted_probe)
        # No evaluation in this info
        card = _render_model_card(config, info)

        assert "No evaluation results provided" in card

    def test_model_card_with_evaluation(self, fitted_probe):
        from lmprobe.hub import _build_probe_config, _build_training_info, _render_model_card

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        metrics = {"accuracy": 0.95, "auroc": 0.98, "n_eval": 20}
        info = _build_training_info(fitted_probe, metrics=metrics)
        card = _render_model_card(config, info)

        assert "0.9500" in card
        assert "0.9800" in card


    def test_model_card_uses_repo_id_in_usage(self, fitted_probe):
        """Issue #57: usage example should show actual repo_id, not placeholder."""
        from lmprobe.hub import _build_probe_config, _build_training_info, _render_model_card

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        info = _build_training_info(fitted_probe)
        card = _render_model_card(
            config, info, repo_id="latent-lab/my-probe"
        )

        assert 'from_hub("latent-lab/my-probe"' in card
        assert "REPO_ID" not in card

    def test_model_card_fallback_repo_id(self, fitted_probe):
        """When repo_id is not provided, falls back to REPO_ID placeholder."""
        from lmprobe.hub import _build_probe_config, _build_training_info, _render_model_card

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        info = _build_training_info(fitted_probe)
        card = _render_model_card(config, info)

        assert "REPO_ID" in card

    def test_model_card_omits_limitations_by_default(self, fitted_probe):
        """Issue #58: no placeholder limitations text when not provided."""
        from lmprobe.hub import _build_probe_config, _build_training_info, _render_model_card

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        info = _build_training_info(fitted_probe)
        card = _render_model_card(config, info)

        assert "Limitations and Intended Use" not in card
        assert "Please fill in" not in card

    def test_model_card_with_limitations(self, fitted_probe):
        """Issue #58: limitations text renders when provided."""
        from lmprobe.hub import _build_probe_config, _build_training_info, _render_model_card

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        info = _build_training_info(fitted_probe)
        card = _render_model_card(
            config, info, limitations="Not suitable for production use."
        )

        assert "## Limitations and Intended Use" in card
        assert "Not suitable for production use." in card

    def test_model_card_compact_layers_for_named_spec(self, fitted_probe):
        """Issue #59: named layer specs show compact format."""
        from lmprobe.hub import _build_probe_config, _build_training_info, _render_model_card

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        # Simulate "all" spec with many layers
        config["probe"]["layers_spec_original"] = "all"
        config["probe"]["layers"] = list(range(30))
        info = _build_training_info(fitted_probe)
        card = _render_model_card(config, info)

        assert "all (0\u201329, 30 layers)" in card
        # Should NOT contain the full list
        assert "[0, 1, 2, 3," not in card

    def test_model_card_eval_hash(self, fitted_probe):
        """Issue #61: eval hash and size shown in model card."""
        from lmprobe.hub import _build_probe_config, _build_training_info, _render_model_card

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        metrics = {
            "accuracy": 0.95,
            "n_eval": 300,
            "eval_hash": "sha256:abc123",
        }
        info = _build_training_info(fitted_probe, metrics=metrics)
        card = _render_model_card(config, info)

        assert "**Evaluation samples**: 300" in card
        assert "**Evaluation hash**: `sha256:abc123`" in card


class TestSerializationRoundtrip:
    """Test classifier serialization and deserialization."""

    @requires_skops
    def test_skops_roundtrip(self, fitted_probe, tmp_path):
        from lmprobe.hub import _load_classifier, _serialize_classifier

        fmt = _serialize_classifier(fitted_probe.classifier_, tmp_path / "classifier")
        assert fmt == "skops"

        loaded = _load_classifier(tmp_path, fmt, trust_classifier=True)
        assert type(loaded).__name__ == type(fitted_probe.classifier_).__name__

        # Verify predictions match
        # Create some test data
        rng = np.random.RandomState(42)
        X = rng.randn(5, fitted_probe.classifier_.coef_.shape[1])
        original_preds = fitted_probe.classifier_.predict(X)
        loaded_preds = loaded.predict(X)
        np.testing.assert_array_equal(original_preds, loaded_preds)

    def test_serialize_roundtrip(self, fitted_probe, tmp_path):
        """Test serialization roundtrip with whatever format is available."""
        from lmprobe.hub import _load_classifier, _serialize_classifier

        fmt = _serialize_classifier(fitted_probe.classifier_, tmp_path / "classifier")
        assert fmt in ("skops", "joblib")

        loaded = _load_classifier(tmp_path, fmt, trust_classifier=True)
        assert type(loaded).__name__ == type(fitted_probe.classifier_).__name__

        rng = np.random.RandomState(42)
        X = rng.randn(5, fitted_probe.classifier_.coef_.shape[1])
        original_preds = fitted_probe.classifier_.predict(X)
        loaded_preds = loaded.predict(X)
        np.testing.assert_array_equal(original_preds, loaded_preds)

    def test_load_without_trust_raises(self, fitted_probe, tmp_path):
        from lmprobe.hub import _load_classifier, _serialize_classifier

        fmt = _serialize_classifier(fitted_probe.classifier_, tmp_path / "classifier")

        with pytest.raises(ValueError, match="trust_classifier=True"):
            _load_classifier(tmp_path, fmt, trust_classifier=False)

    def test_safetensors_not_implemented(self, tmp_path):
        from lmprobe.hub import _load_classifier

        with pytest.raises(NotImplementedError, match="safetensors"):
            _load_classifier(tmp_path, "safetensors", trust_classifier=True)

    def test_unknown_format_raises(self, tmp_path):
        from lmprobe.hub import _load_classifier

        with pytest.raises(ValueError, match="Unknown serialization format"):
            _load_classifier(tmp_path, "unknown_format", trust_classifier=True)


class TestProbeCard:
    """Test ProbeCard dataclass."""

    def test_from_local(self, fitted_probe, tmp_path):
        from lmprobe.hub import (
            ProbeCard,
            _build_probe_config,
            _build_training_info,
        )

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        with open(tmp_path / "probe_config.json", "w") as f:
            json.dump(config, f)

        info = _build_training_info(fitted_probe)
        with open(tmp_path / "training_info.json", "w") as f:
            json.dump(info, f)

        card = ProbeCard.from_local(tmp_path)

        assert card.base_model == TEST_MODEL
        assert card.pooling == "last_token"
        assert card.classifier_type == "logistic_regression"
        assert card.task == "classification"
        assert card.random_state == 42
        assert card.n_positive == 2
        assert card.n_negative == 2
        assert card.positive_hash is not None

    def test_from_local_without_training_info(self, fitted_probe, tmp_path):
        from lmprobe.hub import ProbeCard, _build_probe_config

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        with open(tmp_path / "probe_config.json", "w") as f:
            json.dump(config, f)

        card = ProbeCard.from_local(tmp_path)

        assert card.base_model == TEST_MODEL
        assert card.n_positive is None
        assert card.metrics is None

    def test_to_reproduce_config(self, fitted_probe, tmp_path):
        from lmprobe.hub import ProbeCard, _build_probe_config

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        with open(tmp_path / "probe_config.json", "w") as f:
            json.dump(config, f)

        card = ProbeCard.from_local(tmp_path)
        repro = card.to_reproduce_config()

        assert repro["model"] == TEST_MODEL
        assert repro["pooling"] == "last_token"
        assert repro["classifier"] == "logistic_regression"
        assert repro["random_state"] == 42

    def test_is_compatible_with(self, fitted_probe, tmp_path):
        from lmprobe.hub import ProbeCard, _build_probe_config

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        with open(tmp_path / "probe_config.json", "w") as f:
            json.dump(config, f)

        card = ProbeCard.from_local(tmp_path)
        assert card.is_compatible_with(TEST_MODEL)
        assert not card.is_compatible_with("some-other-model")

    def test_training_data_hash(self, fitted_probe, tmp_path):
        from lmprobe.hub import (
            ProbeCard,
            _build_probe_config,
            _build_training_info,
        )

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = "skops"
        with open(tmp_path / "probe_config.json", "w") as f:
            json.dump(config, f)

        info = _build_training_info(fitted_probe)
        with open(tmp_path / "training_info.json", "w") as f:
            json.dump(info, f)

        card = ProbeCard.from_local(tmp_path)
        assert card.training_data_hash is not None
        assert "sha256:" in card.training_data_hash


class TestPushToHubValidation:
    """Test push_to_hub validation (no network)."""

    def test_push_unfitted_raises(self, tiny_model):
        probe = LinearProbe(
            model=tiny_model,
            layers=-1,
            device="cpu",
            remote=False,
        )
        with pytest.raises(RuntimeError, match="not been fitted"):
            probe.push_to_hub("test/repo")


class TestFromHubValidation:
    """Test from_hub validation (no network)."""

    def test_from_hub_without_trust_raises(self):
        with pytest.raises((ValueError, Exception)):
            LinearProbe.from_hub("nonexistent/repo", trust_classifier=False)


class TestHashPrompts:
    """Test prompt hashing."""

    def test_hash_deterministic(self):
        from lmprobe.hub import _hash_prompts

        h1 = _hash_prompts(["a", "b", "c"])
        h2 = _hash_prompts(["a", "b", "c"])
        assert h1 == h2

    def test_hash_order_invariant(self):
        from lmprobe.hub import _hash_prompts

        h1 = _hash_prompts(["a", "b", "c"])
        h2 = _hash_prompts(["c", "a", "b"])
        assert h1 == h2

    def test_hash_different_for_different_data(self):
        from lmprobe.hub import _hash_prompts

        h1 = _hash_prompts(["a", "b"])
        h2 = _hash_prompts(["x", "y"])
        assert h1 != h2

    def test_hash_starts_with_sha256(self):
        from lmprobe.hub import _hash_prompts

        h = _hash_prompts(["test"])
        assert h.startswith("sha256:")


class TestLocalRoundtrip:
    """Test full serialization roundtrip without network."""

    def test_serialize_and_load_locally(self, fitted_probe, tmp_path):
        """Test the full serialize -> load cycle using local files."""
        from lmprobe.hub import (
            ProbeCard,
            _build_probe_config,
            _build_training_info,
            _load_classifier,
            _render_model_card,
            _serialize_classifier,
        )

        # Serialize
        fmt = _serialize_classifier(fitted_probe.classifier_, tmp_path / "classifier")

        config = _build_probe_config(fitted_probe)
        config["probe"]["serialization_format"] = fmt

        with open(tmp_path / "probe_config.json", "w") as f:
            json.dump(config, f)

        info = _build_training_info(fitted_probe, include_training_data=True)
        with open(tmp_path / "training_info.json", "w") as f:
            json.dump(info, f)

        card_text = _render_model_card(config, info)
        with open(tmp_path / "README.md", "w") as f:
            f.write(card_text)

        # Load and verify
        loaded_classifier = _load_classifier(tmp_path, fmt, trust_classifier=True)

        rng = np.random.RandomState(42)
        X = rng.randn(5, fitted_probe.classifier_.coef_.shape[1])
        original_preds = fitted_probe.classifier_.predict_proba(X)
        loaded_preds = loaded_classifier.predict_proba(X)
        np.testing.assert_array_almost_equal(original_preds, loaded_preds)

        # Verify ProbeCard reads correctly
        card = ProbeCard.from_local(tmp_path)
        assert card.base_model == TEST_MODEL
        assert card.n_positive == 2

        # Verify README
        assert Path(tmp_path / "README.md").exists()
        readme = (tmp_path / "README.md").read_text()
        assert "lmprobe" in readme



"""Tests for LinearProbe.sweep_layers() per-layer probe sweep."""

from lmprobe import LayerSweepResult, LinearProbe


class TestSweepLayers:
    """Tests for the sweep_layers classmethod."""

    def test_sweep_returns_result(self, tiny_model):
        """sweep_layers returns a LayerSweepResult."""
        result = LinearProbe.sweep_layers(
            model=tiny_model,
            positive_prompts=["good", "great"],
            negative_prompts=["bad", "terrible"],
            layers="all",
            pooling="last_token",
            classifier="logistic_regression",
            device="cpu",
            remote=False,
            random_state=42,
        )
        assert isinstance(result, LayerSweepResult)

    def test_sweep_has_all_layers(self, tiny_model):
        """sweep_layers creates a probe for each requested layer."""
        from lmprobe.extraction import get_num_layers_from_config

        num_layers = get_num_layers_from_config(tiny_model)

        result = LinearProbe.sweep_layers(
            model=tiny_model,
            positive_prompts=["good", "great"],
            negative_prompts=["bad", "terrible"],
            layers="all",
            device="cpu",
            remote=False,
            random_state=42,
        )
        assert len(result) == num_layers
        assert result.layers == list(range(num_layers))

    def test_sweep_specific_layers(self, tiny_model):
        """sweep_layers works with specific layer indices."""
        result = LinearProbe.sweep_layers(
            model=tiny_model,
            positive_prompts=["good", "great"],
            negative_prompts=["bad", "terrible"],
            layers=[0, -1],
            device="cpu",
            remote=False,
            random_state=42,
        )
        assert len(result) >= 2
        assert 0 in result.probes

    def test_sweep_single_layer(self, tiny_model):
        """sweep_layers works with a single layer."""
        result = LinearProbe.sweep_layers(
            model=tiny_model,
            positive_prompts=["good", "great"],
            negative_prompts=["bad", "terrible"],
            layers=-1,
            device="cpu",
            remote=False,
            random_state=42,
        )
        assert len(result) == 1

    def test_sweep_probes_are_fitted(self, tiny_model):
        """Each probe in the sweep result is fitted."""
        result = LinearProbe.sweep_layers(
            model=tiny_model,
            positive_prompts=["good", "great"],
            negative_prompts=["bad", "terrible"],
            layers="all",
            device="cpu",
            remote=False,
            random_state=42,
        )
        for layer, probe in result.probes.items():
            assert probe.classifier_ is not None
            assert probe.classes_ is not None

    def test_sweep_score(self, tiny_model):
        """score() returns dict mapping layer -> accuracy."""
        result = LinearProbe.sweep_layers(
            model=tiny_model,
            positive_prompts=["good", "great"],
            negative_prompts=["bad", "terrible"],
            layers="all",
            device="cpu",
            remote=False,
            random_state=42,
        )
        scores = result.score(["test one", "test two"], [1, 0])
        assert isinstance(scores, dict)
        for layer, acc in scores.items():
            assert isinstance(layer, int)
            assert 0.0 <= acc <= 1.0

    def test_sweep_best_layer(self, tiny_model):
        """best_layer() returns an int layer index."""
        result = LinearProbe.sweep_layers(
            model=tiny_model,
            positive_prompts=["good", "great"],
            negative_prompts=["bad", "terrible"],
            layers="all",
            device="cpu",
            remote=False,
            random_state=42,
        )
        best = result.best_layer(["test one", "test two"], [1, 0])
        assert isinstance(best, int)
        assert best in result.probes

    def test_sweep_predict(self, tiny_model):
        """predict() returns dict mapping layer -> predictions."""
        result = LinearProbe.sweep_layers(
            model=tiny_model,
            positive_prompts=["good", "great"],
            negative_prompts=["bad", "terrible"],
            layers="all",
            device="cpu",
            remote=False,
            random_state=42,
        )
        predictions = result.predict(["test one", "test two"])
        assert isinstance(predictions, dict)
        for layer, preds in predictions.items():
            assert preds.shape == (2,)

    def test_sweep_predict_proba(self, tiny_model):
        """predict_proba() returns dict mapping layer -> probabilities."""
        result = LinearProbe.sweep_layers(
            model=tiny_model,
            positive_prompts=["good", "great"],
            negative_prompts=["bad", "terrible"],
            layers="all",
            device="cpu",
            remote=False,
            random_state=42,
        )
        probas = result.predict_proba(["test one", "test two"])
        assert isinstance(probas, dict)
        for layer, proba in probas.items():
            assert proba.shape == (2, 2)

    def test_sweep_getitem(self, tiny_model):
        """LayerSweepResult supports [] indexing by layer."""
        result = LinearProbe.sweep_layers(
            model=tiny_model,
            positive_prompts=["good", "great"],
            negative_prompts=["bad", "terrible"],
            layers=0,
            device="cpu",
            remote=False,
            random_state=42,
        )
        probe = result[0]
        assert isinstance(probe, LinearProbe)

    def test_sweep_each_probe_is_single_layer(self, tiny_model):
        """Each probe in the sweep should use exactly one layer."""
        result = LinearProbe.sweep_layers(
            model=tiny_model,
            positive_prompts=["good", "great"],
            negative_prompts=["bad", "terrible"],
            layers="all",
            device="cpu",
            remote=False,
            random_state=42,
        )
        for layer, probe in result.probes.items():
            assert len(probe._extractor.layer_indices) == 1
            assert probe._extractor.layer_indices[0] == layer
